"""Builtin Speckle object kit."""

from specklepy.objects import (
    GIS,
    encoding,
    geometry,
    other,
    primitive,
    structural,
    units,
)
from specklepy.objects.base import Base

__all__ = [
    "Base",
    "encoding",
    "geometry",
    "other",
    "units",
    "structural",
    "primitive",
    "GIS",
]



import hashlib
import re
import warnings
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple
from uuid import uuid4
from warnings import warn

import ujson

# import for serialization
from specklepy.logging.exceptions import SpeckleException, SpeckleWarning
from specklepy.objects.base import Base, DataChunk
from specklepy.transports.abstract_transport import AbstractTransport

PRIMITIVES = (int, float, str, bool)


def hash_obj(obj: Any) -> str:
    return hashlib.sha256(ujson.dumps(obj).encode()).hexdigest()[:32]


def safe_json_loads(obj: str, obj_id=None) -> Any:
    try:
        return ujson.loads(obj)
    except ValueError as err:
        import json

        warn(
            f"Failed to deserialise object (id: {obj_id}). This is likely a ujson big"
            f" int error - falling back to json. \nError: {err}",
            SpeckleWarning,
        )
        return json.loads(obj)


class BaseObjectSerializer:
    read_transport: AbstractTransport
    write_transports: List[AbstractTransport]
    detach_lineage: List[bool]  # tracks depth and whether or not to detach
    lineage: List[str]  # keeps track of hash chain through the object tree
    family_tree: Dict[str, Dict[str, int]]
    closure_table: Dict[str, Dict[str, int]]
    deserialized: Dict[
        str, Base
    ]  # holds deserialized objects so objects with same id return the same instance

    def __init__(
        self,
        write_transports: Optional[List[AbstractTransport]] = None,
        read_transport: Optional[AbstractTransport] = None,
    ) -> None:
        self.write_transports = write_transports or []
        self.read_transport = read_transport
        self.detach_lineage = []
        self.lineage = []
        self.family_tree = {}
        self.closure_table = {}
        self.deserialized = {}

    def write_json(self, base: Base):
        """Serializes a given base object into a json string
        Arguments:
            base {Base} -- the base object to be decomposed and serialized

        Returns:
            (str, str) -- a tuple containing the object id of the base object and
            the serialized object string
        """

        obj_id, obj = self.traverse_base(base)

        return obj_id, ujson.dumps(obj)

    def traverse_base(self, base: Base) -> Tuple[str, Dict[str, Any]]:
        """Decomposes the given base object and builds a serializable dictionary

        Arguments:
            base {Base} -- the base object to be decomposed and serialized

        Returns:
            (str, dict) -- a tuple containing the object id of the base object and
            the constructed serializable dictionary
        """
        self.__reset_writer()

        if self.write_transports:
            for wt in self.write_transports:
                wt.begin_write()

        obj_id, obj = self._traverse_base(base)

        if self.write_transports:
            for wt in self.write_transports:
                wt.end_write()

        return obj_id, obj

    def _traverse_base(self, base: Base) -> Tuple[str, Dict]:
        if not self.detach_lineage:
            self.detach_lineage = [True]

        self.lineage.append(uuid4().hex)
        object_builder = {"id": "", "speckle_type": "Base", "totalChildrenCount": 0}
        object_builder.update(speckle_type=base.speckle_type)
        obj, props = base, base.get_serializable_attributes()

        while props:
            prop = props.pop(0)
            value = getattr(obj, prop, None)
            chunkable = False
            detach = False

            # skip props marked to be ignored with "__" or "_"
            if prop.startswith(("__", "_")):
                continue

            # don't prepopulate id as this will mess up hashing
            if prop == "id":
                continue

            # only bother with chunking and detaching if there is a write transport
            if self.write_transports:
                dynamic_chunk_match = prop.startswith("@") and re.match(
                    r"^@\((\d*)\)", prop
                )
                if dynamic_chunk_match:
                    chunk_size = dynamic_chunk_match.groups()[0]
                    base._chunkable[prop] = (
                        int(chunk_size) if chunk_size else base._chunk_size_default
                    )

                chunkable = prop in base._chunkable
                detach = bool(
                    prop.startswith("@") or prop in base._detachable or chunkable
                )

            # 1. handle None and primitives (ints, floats, strings, and bools)
            if value is None or isinstance(value, PRIMITIVES):
                object_builder[prop] = value
                continue

            # NOTE: for dynamic props, this won't be re-serialised as an enum but as an int
            if isinstance(value, Enum):
                object_builder[prop] = value.value
                continue

            # 2. handle Base objects
            elif isinstance(value, Base):
                child_obj = self.traverse_value(value, detach=detach)
                if detach and self.write_transports:
                    ref_id = child_obj["id"]
                    object_builder[prop] = self.detach_helper(ref_id=ref_id)
                else:
                    object_builder[prop] = child_obj

            # 3. handle chunkable props
            elif chunkable and self.write_transports:
                chunks = []
                max_size = base._chunkable[prop]
                chunk = DataChunk()
                for count, item in enumerate(value):
                    if count and count % max_size == 0:
                        chunks.append(chunk)
                        chunk = DataChunk()
                    chunk.data.append(item)
                chunks.append(chunk)

                chunk_refs = []
                for c in chunks:
                    self.detach_lineage.append(detach)
                    ref_id, _ = self._traverse_base(c)
                    ref_obj = self.detach_helper(ref_id=ref_id)
                    chunk_refs.append(ref_obj)
                object_builder[prop] = chunk_refs

            # 4. handle all other cases
            else:
                child_obj = self.traverse_value(value, detach)
                object_builder[prop] = child_obj

        closure = {}
        # add closures & children count to the object
        detached = self.detach_lineage.pop()
        if self.lineage[-1] in self.family_tree:
            closure = {
                ref: depth - len(self.detach_lineage)
                for ref, depth in self.family_tree[self.lineage[-1]].items()
            }
        object_builder["totalChildrenCount"] = len(closure)

        obj_id = hash_obj(object_builder)

        object_builder["id"] = obj_id
        if closure:
            object_builder["__closure"] = self.closure_table[obj_id] = closure

        # write detached or root objects to transports
        if detached and self.write_transports:
            for t in self.write_transports:
                t.save_object(id=obj_id, serialized_object=ujson.dumps(object_builder))

        del self.lineage[-1]

        return obj_id, object_builder

    def traverse_value(self, obj: Any, detach: bool = False) -> Any:
        """Decomposes a given object and constructs a serializable object or dictionary

        Arguments:
            obj {Any} -- the value to decompose

        Returns:
            Any -- a serializable version of the given object
        """
        if obj is None:
            return None
        if isinstance(obj, PRIMITIVES):
            return obj

        # NOTE: for dynamic props, this won't be re-serialised as an enum but as an int
        if isinstance(obj, Enum):
            return obj.value

        elif isinstance(obj, (list, tuple, set)):
            if not detach:
                return [self.traverse_value(o) for o in obj]

            detached_list = []
            for o in obj:
                if isinstance(o, Base):
                    self.detach_lineage.append(detach)
                    ref_id, _ = self._traverse_base(o)
                    detached_list.append(self.detach_helper(ref_id=ref_id))
                else:
                    detached_list.append(self.traverse_value(o, detach))
            return detached_list

        elif isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, PRIMITIVES) or v is None:
                    continue
                else:
                    obj[k] = self.traverse_value(v)
            return obj

        elif isinstance(obj, Base):
            self.detach_lineage.append(detach)
            _, base_obj = self._traverse_base(obj)
            return base_obj

        else:
            try:
                return obj.dict()
            except Exception:
                warn(
                    f"Failed to handle {type(obj)} in"
                    " `BaseObjectSerializer.traverse_value`",
                    SpeckleWarning,
                )

                return str(obj)

    def detach_helper(self, ref_id: str) -> Dict[str, str]:
        """
        Helper to keep track of detached objects and their depth in the family tree
        and create reference objects to place in the parent object

        Arguments:
            ref_id {str} -- the id of the fully traversed object

        Returns:
            dict -- a reference object to be inserted into the given object's parent
        """

        for parent in self.lineage:
            if parent not in self.family_tree:
                self.family_tree[parent] = {}
            if ref_id not in self.family_tree[parent] or self.family_tree[parent][
                ref_id
            ] > len(self.detach_lineage):
                self.family_tree[parent][ref_id] = len(self.detach_lineage)

        return {
            "referencedId": ref_id,
            "speckle_type": "reference",
        }

    def __reset_writer(self) -> None:
        """
        Reinitializes the lineage, and other variables that get used during the json
        writing process
        """
        self.detach_lineage = [True]
        self.lineage = []
        self.family_tree = {}
        self.closure_table = {}

    def read_json(self, obj_string: str) -> Base:
        """Recomposes a Base object from the string representation of the object

        Arguments:
            obj_string {str} -- the string representation of the object

        Returns:
            Base -- the base object with all it's children attached
        """
        if not obj_string:
            return None

        self.deserialized = {}
        obj = safe_json_loads(obj_string)
        return self.recompose_base(obj=obj)

    def recompose_base(self, obj: dict) -> Base:
        """Steps through a base object dictionary and recomposes the base object

        Arguments:
            obj {dict} -- the dictionary representation of the object

        Returns:
            Base -- the base object with all its children attached
        """
        # make sure an obj was passed and create dict if string was somehow passed
        if not obj:
            return
        if isinstance(obj, str):
            obj = safe_json_loads(obj)

        if "id" in obj and obj["id"] in self.deserialized:
            return self.deserialized[obj["id"]]

        if "speckle_type" in obj and obj["speckle_type"] == "reference":
            obj = self.get_child(obj=obj)

        speckle_type = obj.get("speckle_type")
        # if speckle type is not in the object definition, it is treated as a dict
        if not speckle_type:
            return obj

        # get the registered type from base register.
        object_type = Base.get_registered_type(speckle_type)

        # initialise the base object using `speckle_type` fall back to base if needed
        base = object_type() if object_type else Base.of_type(speckle_type=speckle_type)
        # get total children count
        if "__closure" in obj:
            if not self.read_transport:
                raise SpeckleException(
                    message="Cannot resolve reference - no read transport is defined"
                )
            closure = obj.pop("__closure")
            base.totalChildrenCount = len(closure)

        for prop, value in obj.items():
            # 1. handle primitives (ints, floats, strings, and bools) or None
            if isinstance(value, PRIMITIVES) or value is None:
                base.__setattr__(prop, value)
                continue

            # 2. handle referenced child objects
            elif "referencedId" in value:
                ref_id = value["referencedId"]
                ref_obj_str = self.read_transport.get_object(id=ref_id)
                if ref_obj_str:
                    ref_obj = safe_json_loads(ref_obj_str, ref_id)
                    base.__setattr__(prop, self.recompose_base(obj=ref_obj))
                else:
                    warnings.warn(
                        f"Could not find the referenced child object of id `{ref_id}`"
                        f" in the given read transport: {self.read_transport.name}",
                        SpeckleWarning,
                    )
                    base.__setattr__(prop, self.handle_value(value))

            # 3. handle all other cases (base objects, lists, and dicts)
            else:
                base.__setattr__(prop, self.handle_value(value))

        if "id" in obj:
            self.deserialized[obj["id"]] = base

        return base

    def handle_value(self, obj: Any):
        """Helper for recomposing a base object by handling the dictionary
        representation's values

        Arguments:
            obj {Any} -- a value from the base object dictionary

        Returns:
            Any -- the handled value (primitive, list, dictionary, or Base)
        """
        if not obj:
            return obj

        if isinstance(obj, PRIMITIVES):
            return obj

        # lists (regular and chunked)
        if isinstance(obj, list):
            obj_list = [self.handle_value(o) for o in obj]
            if (
                hasattr(obj_list[0], "speckle_type")
                and "DataChunk" in obj_list[0].speckle_type
            ):
                # handle chunked lists
                data = []
                for o in obj_list:
                    data.extend(o.data)
                return data
            return obj_list

        # bases
        if isinstance(obj, dict) and "speckle_type" in obj:
            return self.recompose_base(obj=obj)

        # dictionaries
        if isinstance(obj, dict):
            for k, v in obj.items():
                if isinstance(v, PRIMITIVES):
                    continue
                else:
                    obj[k] = self.handle_value(v)
            return obj

    def get_child(self, obj: Dict):
        ref_id = obj["referencedId"]
        ref_obj_str = self.read_transport.get_object(id=ref_id)
        if not ref_obj_str:
            warnings.warn(
                f"Could not find the referenced child object of id `{ref_id}` in the"
                f" given read transport: {self.read_transport.name}",
                SpeckleWarning,
            )
            return obj

        return safe_json_loads(ref_obj_str, ref_id)



from typing import List, Optional, Union

from specklepy.objects.base import Base
from specklepy.objects.geometry import (
    Arc,
    Circle,
    Line,
    Mesh,
    Point,
    Polycurve,
    Polyline,
)


class PolygonGeometry(Base, speckle_type="Objects.GIS.PolygonGeometry"):
    """GIS Polygon Geometry"""

    boundary: Optional[Polyline]
    voids: Optional[List[Polyline]]


GisPolygonGeometry = PolygonGeometry


class GisPolygonElement(Base, speckle_type="Objects.GIS.PolygonElement"):
    """GIS Polygon element"""

    geometry: Optional[List[GisPolygonGeometry]] = None
    attributes: Optional[Base] = None


class GisLineElement(Base, speckle_type="Objects.GIS.LineElement"):
    """GIS Polyline element"""

    geometry: Optional[List[Union[Polyline, Arc, Line, Circle, Polycurve]]] = None
    attributes: Optional[Base] = None


class GisPointElement(Base, speckle_type="Objects.GIS.PointElement"):
    """GIS Point element"""

    geometry: Optional[List[Point]] = None
    attributes: Optional[Base] = None


class GisRasterElement(
    Base, speckle_type="Objects.GIS.RasterElement", detachable={"displayValue"}
):
    """GIS Raster element"""

    band_count: Optional[int] = None
    band_names: Optional[List[str]] = None
    x_origin: Optional[float] = None
    y_origin: Optional[float] = None
    x_size: Optional[int] = None
    y_size: Optional[int] = None
    x_resolution: Optional[float] = None
    y_resolution: Optional[float] = None
    noDataValue: Optional[List[float]] = None
    displayValue: Optional[List[Mesh]] = None


class GisTopography(
    GisRasterElement,
    speckle_type="Objects.GIS.GisTopography",
    detachable={"displayValue"},
):
    """GIS Raster element with 3d Topography representation"""


class GisNonGeometryElement(Base, speckle_type="Objects.GIS.NonGeometryElement"):
    """GIS Table feature"""

    attributes: Optional[Base] = None



from functools import wraps
from typing import Callable, Dict, List, Optional, Union

from deprecated import deprecated
from gql import gql
from graphql import DocumentNode

from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
    Stream,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.logging.exceptions import SpeckleException

NAME = "subscribe"


def check_wsclient(function):
    @wraps(function)
    async def check_wsclient_wrapper(self, *args, **kwargs):
        if self.client is None:
            raise SpeckleException(
                "You must authenticate before you can subscribe to events"
            )
        else:
            return await function(self, *args, **kwargs)

    return check_wsclient_wrapper


class Resource(ResourceBase):
    """API Access class for subscriptions"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    @check_wsclient
    async def stream_added(self, callback: Optional[Callable] = None):
        """Subscribes to new stream added event for your profile.
        Use this to display an up-to-date list of streams.

        Arguments:
            callback {Callable[Stream]} -- a function that takes the updated stream
            as an argument and executes each time a stream is added

        Returns:
            Stream -- the update stream
        """
        query = gql(
            """
            subscription { userStreamAdded }
            """
        )
        return await self.subscribe(
            query=query, callback=callback, return_type="userStreamAdded", schema=Stream
        )

    @check_wsclient
    async def stream_updated(self, id: str, callback: Optional[Callable] = None):
        """
        Subscribes to stream updated event.
        Use this in clients/components that pertain only to this stream.

        Arguments:
            id {str} -- the stream id of the stream to subscribe to
            callback {Callable[Stream]}
            -- a function that takes the updated stream
            as an argument and executes each time the stream is updated

        Returns:
            Stream -- the update stream
        """
        query = gql(
            """
            subscription Update($id: String!) { streamUpdated(streamId: $id) }
            """
        )
        params = {"id": id}

        return await self.subscribe(
            query=query,
            params=params,
            callback=callback,
            return_type="streamUpdated",
            schema=Stream,
        )

    @check_wsclient
    async def stream_removed(self, callback: Optional[Callable] = None):
        """Subscribes to stream removed event for your profile.
        Use this to display an up-to-date list of streams for your profile.
        NOTE: If someone revokes your permissions on a stream,
        this subscription will be triggered with an extra value of revokedBy
        in the payload.

        Arguments:
            callback {Callable[Dict]}
            -- a function that takes the returned dict as an argument
            and executes each time a stream is removed

        Returns:
            dict -- dict containing 'id' of stream removed and optionally 'revokedBy'
        """
        query = gql(
            """
            subscription { userStreamRemoved }
            """
        )

        return await self.subscribe(
            query=query,
            callback=callback,
            return_type="userStreamRemoved",
            parse_response=False,
        )

    @check_wsclient
    async def subscribe(
        self,
        query: DocumentNode,
        params: Optional[Dict] = None,
        callback: Optional[Callable] = None,
        return_type: Optional[Union[str, List]] = None,
        schema=None,
        parse_response: bool = True,
    ):
        # if self.client.transport.websocket is None:
        # TODO: add multiple subs to the same ws connection
        async with self.client as session:
            async for res in session.subscribe(query, variable_values=params):
                res = self._step_into_response(response=res, return_type=return_type)
                if parse_response:
                    res = self._parse_response(response=res, schema=schema)
                if callback is not None:
                    callback(res)
                else:
                    return res



from functools import wraps
from typing import Any, Callable, Dict, Optional, Sequence, Type

from gql import gql
from graphql import DocumentNode
from pydantic import BaseModel
from typing_extensions import TypeVar

from specklepy.core.api.models import (
    ProjectModelsUpdatedMessage,
    ProjectUpdatedMessage,
    ProjectVersionsUpdatedMessage,
    UserProjectsUpdatedMessage,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse
from specklepy.logging.exceptions import SpeckleException

NAME = "subscribe"

TEventArgs = TypeVar("TEventArgs", bound=BaseModel)


def check_wsclient(function):
    @wraps(function)
    async def check_wsclient_wrapper(self, *args, **kwargs):
        if self.client is None:
            raise SpeckleException(
                "You must authenticate before you can subscribe to events"
            )
        else:
            return await function(self, *args, **kwargs)

    return check_wsclient_wrapper


class SubscriptionResource(ResourceBase):
    """API Access class for subscriptions"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
        )

    async def user_projects_updated(
        self, callback: Callable[[UserProjectsUpdatedMessage], None]
    ) -> None:
        QUERY = gql(
            """
            subscription UserProjectsUpdated {
              data:userProjectsUpdated {
                id
                project {
                  id
                  name
                  description
                  visibility
                  allowPublicComments
                  role
                  createdAt
                  updatedAt
                  sourceApps
                  workspaceId
                }
                type
              }
            }
            """
        )

        await self.subscribe_2(
            DataResponse[UserProjectsUpdatedMessage],
            QUERY,
            None,
            callback=lambda d: callback(d.data),
        )

    async def project_models_updated(
        self,
        callback: Callable[[ProjectModelsUpdatedMessage], None],
        id: str,
        model_ids: Optional[Sequence[str]] = None,
    ) -> None:
        QUERY = gql(
            """
            subscription ProjectModelsUpdated($id: String!, $modelIds: [String!]) {
              data:projectModelsUpdated(id: $id, modelIds: $modelIds) {
                id
                model {
                  id
                  name
                  previewUrl
                  updatedAt
                  description
                  displayName
                  createdAt
                  author {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                }
                type
              }
            }
            """
        )

        variables = {"id": id, "modelIds": model_ids}

        await self.subscribe_2(
            DataResponse[ProjectModelsUpdatedMessage],
            QUERY,
            variables,
            callback=lambda d: callback(d.data),
        )

    async def project_updated(
        self,
        callback: Callable[[ProjectUpdatedMessage], None],
        id: str,
    ) -> None:
        QUERY = gql(
            """
            subscription ProjectUpdated($id: String!) {
              data:projectUpdated(id: $id) {
                id
                project {
                  id
                  name
                  description
                  visibility
                  allowPublicComments
                  role
                  createdAt
                  updatedAt
                  sourceApps
                  workspaceId
                }
                type
              }
            }
            """
        )

        variables = {"id": id}

        await self.subscribe_2(
            DataResponse[ProjectUpdatedMessage],
            QUERY,
            variables,
            callback=lambda d: callback(d.data),
        )

    async def project_versions_updated(
        self,
        callback: Callable[[ProjectVersionsUpdatedMessage], None],
        id: str,
    ) -> None:
        QUERY = gql(
            """
            subscription ProjectVersionsUpdated($id: String!) {
              data:projectVersionsUpdated(id: $id) {
                id
                modelId
                type
                version {
                  id
                  referencedObject
                  message
                  sourceApplication
                  createdAt
                  previewUrl
                  authorUser {
                    id
                    name
                    bio
                    company
                    verified
                    role
                    avatar
                  }
                }
              }
            }
            """
        )

        variables = {"id": id}

        await self.subscribe_2(
            DataResponse[ProjectVersionsUpdatedMessage],
            QUERY,
            variables,
            callback=lambda d: callback(d.data),
        )

    @check_wsclient
    async def subscribe_2(
        self,
        response_type: Type[TEventArgs],
        query: DocumentNode,
        variables: Optional[Dict[str, Any]],
        callback: Callable[[TEventArgs], None],
    ) -> None:
        async with self.client as session:
            self.session = session
            gen = session.subscribe(query, variable_values=variables)
            async for res in gen:
                event_arg = response_type.model_validate(res)
                callback(event_arg)



from typing import Any, Optional, Tuple

from gql import Client

from specklepy.core.api.credentials import Account
from specklepy.core.api.inputs.project_inputs import (
    ProjectInviteCreateInput,
    ProjectInviteUseInput,
)
from specklepy.core.api.models import PendingStreamCollaborator, ProjectWithTeam
from specklepy.core.api.resources import ProjectInviteResource as CoreResource
from specklepy.logging import metrics


class ProjectInviteResource(CoreResource):
    """API Access class for project invites"""

    def __init__(
        self,
        account: Account,
        basepath: str,
        client: Client,
        server_version: Optional[Tuple[Any, ...]],
    ) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )

    def create(
        self, project_id: str, input: ProjectInviteCreateInput
    ) -> ProjectWithTeam:
        metrics.track(metrics.SDK, self.account, {"name": "Project Invite Create"})
        return super().create(project_id, input)

    def use(self, input: ProjectInviteUseInput) -> bool:
        metrics.track(metrics.SDK, self.account, {"name": "Project Invite Use"})
        return super().use(input)

    def get(
        self, project_id: str, token: Optional[str]
    ) -> Optional[PendingStreamCollaborator]:
        metrics.track(metrics.SDK, self.account, {"name": "Project Invite Get"})
        return super().get(project_id, token)

    def cancel(
        self,
        project_id: str,
        invite_id: str,
    ) -> ProjectWithTeam:
        metrics.track(metrics.SDK, self.account, {"name": "Project Invite Cancel"})
        return super().cancel(project_id, invite_id)



from specklepy.core.api.resources.current.active_user_resource import ActiveUserResource
from specklepy.core.api.resources.current.model_resource import ModelResource
from specklepy.core.api.resources.current.other_user_resource import OtherUserResource
from specklepy.core.api.resources.current.project_invite_resource import (
    ProjectInviteResource,
)
from specklepy.core.api.resources.current.project_resource import ProjectResource
from specklepy.core.api.resources.current.server_resource import ServerResource
from specklepy.core.api.resources.current.subscription_resource import (
    SubscriptionResource,
)
from specklepy.core.api.resources.current.version_resource import VersionResource
from specklepy.core.api.resources.deprecated import (
    active_user,
    branch,
    commit,
    object,
    other_user,
    server,
    stream,
    subscriptions,
    user,
)

__all__ = [
    "ActiveUserResource",
    "ModelResource",
    "OtherUserResource",
    "ProjectInviteResource",
    "ProjectResource",
    "ServerResource",
    "SubscriptionResource",
    "VersionResource",
    "active_user",
    "branch",
    "commit",
    "object",
    "other_user",
    "server",
    "stream",
    "subscriptions",
    "user",
]



""""""

from enum import Enum
from typing import Any, Dict, List, Literal, Optional

from pydantic import BaseModel, ConfigDict, Field
from stringcase import camelcase


class AutomateBase(BaseModel):
    """Use this class as a base model for automate related DTO."""

    model_config = ConfigDict(alias_generator=camelcase, populate_by_name=True)


class VersionCreationTriggerPayload(AutomateBase):
    """Represents the version creation trigger payload."""

    model_id: str
    version_id: str


class VersionCreationTrigger(AutomateBase):
    """Represents a single version creation trigger for the automation run."""

    trigger_type: Literal["versionCreation"]
    payload: VersionCreationTriggerPayload


class AutomationRunData(BaseModel):
    """Values of the project / model that triggered the run of this function."""

    project_id: str
    speckle_server_url: str
    automation_id: str
    automation_run_id: str
    function_run_id: str

    triggers: List[VersionCreationTrigger]

    model_config = ConfigDict(
        alias_generator=camelcase, populate_by_name=True, protected_namespaces=()
    )


class TestAutomationRunData(BaseModel):
    """Values of the run created in the test automation for local test results."""

    automation_run_id: str
    function_run_id: str

    triggers: List[VersionCreationTrigger]

    model_config = ConfigDict(
        alias_generator=camelcase, populate_by_name=True, protected_namespaces=()
    )


class AutomationStatus(str, Enum):
    """Set the status of the automation."""

    INITIALIZING = "INITIALIZING"
    RUNNING = "RUNNING"
    FAILED = "FAILED"
    SUCCEEDED = "SUCCEEDED"
    EXCEPTION = "EXCEPTION"


class ObjectResultLevel(str, Enum):
    """Possible status message levels for object reports."""

    SUCCESS = "SUCCESS"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"


class ResultCase(AutomateBase):
    """A result case."""

    category: str
    level: ObjectResultLevel
    object_ids: List[str]
    message: Optional[str]
    metadata: Optional[Dict[str, Any]]
    visual_overrides: Optional[Dict[str, Any]]


class AutomationResult(AutomateBase):
    """Schema accepted by the Speckle server as a result for an automation run."""

    elapsed: float = 0
    result_view: Optional[str] = None
    result_versions: List[str] = Field(default_factory=list)
    blobs: List[str] = Field(default_factory=list)
    run_status: AutomationStatus = AutomationStatus.RUNNING
    status_message: Optional[str] = None
    object_results: list[ResultCase] = Field(default_factory=list)



from enum import Enum
from typing import List, Optional

from specklepy.objects.base import Base
from specklepy.objects.geometry import Vector
from specklepy.objects.structural.axis import Axis

STRUCTURAL_LOADING = "Objects.Structural.Loading."


class LoadType(int, Enum):
    none = 0
    Dead = 1
    SuperDead = 2
    Soil = 3
    Live = 4
    LiveRoof = 5
    ReducibleLive = 6
    Wind = 7
    Snow = 8
    Rain = 9
    Thermal = 10
    Notional = 11
    Prestress = 12
    Equivalent = 13
    Accidental = 14
    SeismicRSA = 15
    SeismicAccTorsion = 16
    SeismicStatic = 17
    Other = 18


class ActionType(int, Enum):
    none = 0
    Permanent = 1
    Variable = 2
    Accidental = 3


class BeamLoadType(int, Enum):
    Point = 0
    Uniform = 1
    Linear = 2
    Patch = 3
    TriLinear = 4


class FaceLoadType(int, Enum):
    Constant = 0
    Variable = 1
    Point = 2


class LoadDirection2D(int, Enum):
    X = 0
    Y = 1
    Z = 2


class LoadDirection(int, Enum):
    X = 0
    Y = 1
    Z = 2
    XX = 3
    YY = 4
    ZZ = 5


class LoadAxisType(int, Enum):
    Global = 0
    Local = 1  # local element axes
    DeformedLocal = (
        2  # element local axis that is embedded in the element as it deforms
    )


class CombinationType(int, Enum):
    LinearAdd = 0
    Envelope = 1
    AbsoluteAdd = 2
    SRSS = 3
    RangeAdd = 4


class LoadCase(Base, speckle_type=STRUCTURAL_LOADING + "LoadCase"):
    name: Optional[str] = None
    loadType: Optional[LoadType] = None
    group: Optional[str] = None
    actionType: Optional[ActionType] = None
    description: Optional[str] = None


class Load(Base, speckle_type=STRUCTURAL_LOADING + "Load"):
    name: Optional[str] = None
    loadCase: Optional[LoadCase] = None


class LoadBeam(Load, speckle_type=STRUCTURAL_LOADING + "LoadBeam"):
    elements: Optional[List] = None
    loadType: Optional[BeamLoadType] = None
    direction: Optional[LoadDirection] = None
    loadAxis: Optional[Axis] = None
    loadAxisType: Optional[LoadAxisType] = None
    isProjected: Optional[bool] = None
    values: Optional[List] = None
    positions: Optional[List] = None


class LoadCombinations(Base, speckle_type=STRUCTURAL_LOADING + "LoadCombination"):
    name: Optional[str] = None
    loadCases: List
    loadFactors: List
    combinationType: CombinationType


class LoadFace(Load, speckle_type=STRUCTURAL_LOADING + "LoadFace"):
    elements: Optional[List] = None
    loadType: Optional[FaceLoadType] = None
    direction: Optional[LoadDirection2D] = None
    loadAxis: Optional[Axis] = None
    loadAxisType: Optional[LoadAxisType] = None
    isProjected: Optional[bool] = None
    values: Optional[List] = None
    positions: Optional[List] = None


class LoadGravity(Load, speckle_type=STRUCTURAL_LOADING + "LoadGravity"):
    elements: Optional[List] = None
    nodes: Optional[List] = None
    gravityFactors: Optional[Vector] = None


class LoadNode(Load, speckle_type=STRUCTURAL_LOADING + "LoadNode"):
    nodes: Optional[List] = None
    loadAxis: Optional[Axis] = None
    direction: Optional[LoadDirection] = None
    value: float = 0.0



from typing import List, Optional

# from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException
from specklepy.objects.base import Base
from specklepy.serialization.base_object_serializer import BaseObjectSerializer
from specklepy.transports.abstract_transport import AbstractTransport
from specklepy.transports.sqlite import SQLiteTransport


def send(
    base: Base,
    transports: Optional[List[AbstractTransport]] = None,
    use_default_cache: bool = True,
):
    """Sends an object via the provided transports. Defaults to the local cache.

    Arguments:
        obj {Base} -- the object you want to send
        transports {list} -- where you want to send them
        use_default_cache {bool} -- toggle for the default cache.
        If set to false, it will only send to the provided transports

    Returns:
        str -- the object id of the sent object
    """

    if not transports and not use_default_cache:
        raise SpeckleException(
            message=(
                "You need to provide at least one transport: cannot send with an empty"
                " transport list and no default cache"
            )
        )

    if isinstance(transports, AbstractTransport):
        transports = [transports]

    if transports is None:
        transports = []

    if use_default_cache:
        transports.insert(0, SQLiteTransport())

    serializer = BaseObjectSerializer(write_transports=transports)

    obj_hash, _ = serializer.write_json(base=base)

    return obj_hash


def receive(
    obj_id: str,
    remote_transport: Optional[AbstractTransport] = None,
    local_transport: Optional[AbstractTransport] = None,
) -> Base:
    """Receives an object from a transport.

    Arguments:
        obj_id {str} -- the id of the object to receive
        remote_transport {Transport} -- the transport to receive from
        local_transport {Transport} -- the local cache to check for existing objects
                                       (defaults to `SQLiteTransport`)

    Returns:
        Base -- the base object
    """
    if not local_transport:
        local_transport = SQLiteTransport()

    serializer = BaseObjectSerializer(read_transport=local_transport)

    # try local transport first. if the parent is there, we assume all the children are there and continue with deserialization using the local transport
    obj_string = local_transport.get_object(obj_id)
    if obj_string:
        return serializer.read_json(obj_string=obj_string)

    if not remote_transport:
        raise SpeckleException(
            message=(
                "Could not find the specified object using the local transport, and you"
                " didn't provide a fallback remote from which to pull it."
            )
        )

    obj_string = remote_transport.copy_object_and_children(
        id=obj_id, target_transport=local_transport
    )

    return serializer.read_json(obj_string=obj_string)


def serialize(base: Base, write_transports: List[AbstractTransport] = []) -> str:
    """
    Serialize a base object. If no write transports are provided,
    the object will be serialized
    without detaching or chunking any of the attributes.

    Arguments:
        base {Base} -- the object to serialize
        write_transports {List[AbstractTransport]}
        -- optional: the transports to write to

    Returns:
        str -- the serialized object
    """
    serializer = BaseObjectSerializer(write_transports=write_transports)

    return serializer.write_json(base)[1]


def deserialize(
    obj_string: str, read_transport: Optional[AbstractTransport] = None
) -> Base:
    """
    Deserialize a string object into a Base object.

    If the object contains referenced child objects that are not stored in the local db,
    a read transport needs to be provided in order to recompose
    the base with the children objects.

    Arguments:
        obj_string {str} -- the string object to deserialize
        read_transport {AbstractTransport}
            -- the transport to fetch children objects from
                (defaults to SQLiteTransport)

    Returns:
        Base -- the deserialized object
    """
    if not read_transport:
        read_transport = SQLiteTransport()

    serializer = BaseObjectSerializer(read_transport=read_transport)

    return serializer.read_json(obj_string=obj_string)


__all__ = ["receive", "send", "serialize", "deserialize"]



from typing import Any, Dict, List, Optional, Union

from deprecated import deprecated

from specklepy.objects.base import Base
from specklepy.objects.GIS.CRS import CRS
from specklepy.objects.other import Collection


@deprecated(version="2.15", reason="Use VectorLayer or RasterLayer instead")
class Layer(Base, detachable={"features"}):
    """A GIS Layer"""

    def __init__(
        self,
        name: Optional[str] = None,
        crs: Optional[CRS] = None,
        units: str = "m",
        features: Optional[List[Base]] = None,
        layerType: str = "None",
        geomType: str = "None",
        renderer: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.name = name
        self.crs = crs
        self.units = units
        self.type = layerType
        self.features = features or []
        self.geomType = geomType
        self.renderer = renderer or {}


@deprecated(version="2.16", reason="Use VectorLayer or RasterLayer instead")
class VectorLayer(
    Collection,
    detachable={"elements"},
    speckle_type="VectorLayer",
    serialize_ignore={"features"},
):
    """GIS Vector Layer"""

    name: Optional[str] = None
    crs: Optional[Union[CRS, Base]] = None
    units: Optional[str] = None
    elements: Optional[List[Base]] = None
    attributes: Optional[Base] = None
    geomType: Optional[str] = "None"
    renderer: Optional[Dict[str, Any]] = None
    collectionType = "VectorLayer"

    @property
    @deprecated(version="2.14", reason="Use elements")
    def features(self) -> Optional[List[Base]]:
        return self.elements

    @features.setter
    def features(self, value: Optional[List[Base]]) -> None:
        self.elements = value


@deprecated(version="2.16", reason="Use VectorLayer or RasterLayer instead")
class RasterLayer(
    Collection,
    detachable={"elements"},
    speckle_type="RasterLayer",
    serialize_ignore={"features"},
):
    """GIS Raster Layer"""

    name: Optional[str] = None
    crs: Optional[Union[CRS, Base]] = None
    units: Optional[str] = None
    rasterCrs: Optional[Union[CRS, Base]] = None
    elements: Optional[List[Base]] = None
    geomType: Optional[str] = "None"
    renderer: Optional[Dict[str, Any]] = None
    collectionType = "RasterLayer"

    @property
    @deprecated(version="2.14", reason="Use elements")
    def features(self) -> Optional[List[Base]]:
        return self.elements

    @features.setter
    def features(self, value: Optional[List[Base]]) -> None:
        self.elements = value


class VectorLayer(  # noqa: F811
    Collection,
    detachable={"elements"},
    speckle_type="Objects.GIS.VectorLayer",
    serialize_ignore={"features"},
):
    """GIS Vector Layer"""

    name: Optional[str] = None
    crs: Optional[Union[CRS, Base]] = None
    units: Optional[str] = None
    elements: Optional[List[Base]] = None
    attributes: Optional[Base] = None
    geomType: Optional[str] = "None"
    renderer: Optional[Dict[str, Any]] = None
    collectionType = "VectorLayer"

    @property
    @deprecated(version="2.14", reason="Use elements")
    def features(self) -> Optional[List[Base]]:
        return self.elements

    @features.setter
    def features(self, value: Optional[List[Base]]) -> None:
        self.elements = value


class RasterLayer(  # noqa: F811
    Collection,
    detachable={"elements"},
    speckle_type="Objects.GIS.RasterLayer",
    serialize_ignore={"features"},
):
    """GIS Raster Layer"""

    name: Optional[str] = None
    crs: Optional[Union[CRS, Base]] = None
    units: Optional[str] = None
    rasterCrs: Optional[Union[CRS, Base]] = None
    elements: Optional[List[Base]] = None
    geomType: Optional[str] = "None"
    renderer: Optional[Dict[str, Any]] = None
    collectionType = "RasterLayer"

    @property
    @deprecated(version="2.14", reason="Use elements")
    def features(self) -> Optional[List[Base]]:
        return self.elements

    @features.setter
    def features(self, value: Optional[List[Base]]) -> None:
        self.elements = value



from deprecated import deprecated

from specklepy.api.credentials import Account
from specklepy.api.resources import (
    ActiveUserResource,
    ModelResource,
    OtherUserResource,
    ProjectInviteResource,
    ProjectResource,
    SubscriptionResource,
    VersionResource,
    branch,
    commit,
    object,
    server,
    stream,
    subscriptions,
    user,
)
from specklepy.core.api.client import SpeckleClient as CoreSpeckleClient
from specklepy.logging import metrics


class SpeckleClient(CoreSpeckleClient):
    """
    The `SpeckleClient` is your entry point for interacting with
    your Speckle Server's GraphQL API.
    You'll need to have access to a server to use it,
    or you can use our public server `app.speckle.systems`.

    To authenticate the client, you'll need to have downloaded
    the [Speckle Manager](https://speckle.guide/#speckle-manager)
    and added your account.

    ```py
    from specklepy.api.client import SpeckleClient
    from specklepy.api.credentials import get_default_account

    # initialise the client
    client = SpeckleClient(host="app.speckle.systems") # or whatever your host is
    # client = SpeckleClient(host="localhost:3000", use_ssl=False) or use local server

    # authenticate the client with an account (account has been added in Speckle Manager)
    account = get_default_account()
    client.authenticate_with_account(account)

    # create a new stream. this returns the stream id
    new_stream_id = client.stream.create(name="a shiny new stream")

    # use that stream id to get the stream from the server
    new_stream = client.stream.get(id=new_stream_id)
    ```
    """

    DEFAULT_HOST = "app.speckle.systems"
    USE_SSL = True

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        use_ssl: bool = USE_SSL,
        verify_certificate: bool = True,
    ) -> None:
        super().__init__(
            host=host,
            use_ssl=use_ssl,
            verify_certificate=verify_certificate,
        )
        self.account = Account()

    def _init_resources(self) -> None:
        self.server = server.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )

        server_version = None
        try:
            server_version = self.server.version()
        except Exception:
            pass

        self.other_user = OtherUserResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.active_user = ActiveUserResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.project = ProjectResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.project_invite = ProjectInviteResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.model = ModelResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.version = VersionResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.subscription = SubscriptionResource(
            account=self.account,
            basepath=self.ws_url,
            client=self.wsclient,
            # todo: why doesn't this take a server version
        )
        # Deprecated Resources
        self.user = user.Resource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.stream = stream.Resource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.commit = commit.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )
        self.branch = branch.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )
        self.object = object.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )
        self.subscribe = subscriptions.Resource(
            account=self.account,
            basepath=self.ws_url,
            client=self.wsclient,
        )

    @deprecated(
        version="2.6.0",
        reason=(
            "Renamed: please use `authenticate_with_account` or"
            " `authenticate_with_token` instead."
        ),
    )
    def authenticate(self, token: str) -> None:
        """Authenticate the client using a personal access token
        The token is saved in the client object and a synchronous GraphQL
        entrypoint is created

        Arguments:
            token {str} -- an api token
        """
        metrics.track(
            metrics.SDK, self.account, {"name": "Client Authenticate_deprecated"}
        )
        return super().authenticate(token)

    def authenticate_with_token(self, token: str) -> None:
        """
        Authenticate the client using a personal access token.
        The token is saved in the client object and a synchronous GraphQL
        entrypoint is created

        Arguments:
            token {str} -- an api token
        """
        metrics.track(
            metrics.SDK, self.account, {"name": "Client Authenticate With Token"}
        )
        return super().authenticate_with_token(token)

    def authenticate_with_account(self, account: Account) -> None:
        """Authenticate the client using an Account object
        The account is saved in the client object and a synchronous GraphQL
        entrypoint is created

        Arguments:
            account {Account} -- the account object which can be found with
            `get_default_account` or `get_local_accounts`
        """
        metrics.track(
            metrics.SDK, self.account, {"name": "Client Authenticate With Account"}
        )
        return super().authenticate_with_account(account)



from typing import Any, Optional, Tuple

from gql import Client, gql

from specklepy.core.api.credentials import Account
from specklepy.core.api.inputs.project_inputs import (
    ProjectInviteCreateInput,
    ProjectInviteUseInput,
)
from specklepy.core.api.models import PendingStreamCollaborator, ProjectWithTeam
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse

NAME = "project_invite"


class ProjectInviteResource(ResourceBase):
    """API Access class for project invites"""

    def __init__(
        self,
        account: Account,
        basepath: str,
        client: Client,
        server_version: Optional[Tuple[Any, ...]],
    ) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )

    def create(
        self, project_id: str, input: ProjectInviteCreateInput
    ) -> ProjectWithTeam:
        QUERY = gql(
            """
            mutation ProjectInviteCreate($projectId: ID!, $input: ProjectInviteCreateInput!) {
              data:projectMutations {
                data:invites {
                  data:create(projectId: $projectId, input: $input) {
                    id
                    name
                    description
                    visibility
                    allowPublicComments
                    role
                    createdAt
                    updatedAt
                    workspaceId
                    sourceApps
                    team {
                      id
                      role
                      user {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                      }
                    }
                    invitedTeam {
                      id
                      inviteId
                      projectId
                      projectName
                      title
                      role
                      token
                      user {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                      }
                      invitedBy {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                      }
                    }
                  }
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[DataResponse[ProjectWithTeam]]], QUERY, variables
        ).data.data.data

    def use(self, input: ProjectInviteUseInput) -> bool:
        QUERY = gql(
            """
            mutation ProjectInviteUse($input: ProjectInviteUseInput!) {
              data:projectMutations {
                data:invites {
                  data:use(input: $input)
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[DataResponse[bool]]], QUERY, variables
        ).data.data.data

    def get(
        self, project_id: str, token: Optional[str]
    ) -> Optional[PendingStreamCollaborator]:
        """Returns: The invite, or None if no invite exists"""

        QUERY = gql(
            """
            query ProjectInvite($projectId: String!, $token: String) {
              data:projectInvite(projectId: $projectId, token: $token) {
                id
                inviteId
                invitedBy {
                  avatar
                  bio
                  company
                  id
                  name
                  role
                  verified
                }
                projectId
                projectName
                role
                title
                token
                user {
                  avatar
                  bio
                  company
                  id
                  name
                  role
                  verified
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "token": token,
        }

        return self.make_request_and_parse_response(
            DataResponse[Optional[PendingStreamCollaborator]], QUERY, variables
        ).data

    def cancel(
        self,
        project_id: str,
        invite_id: str,
    ) -> ProjectWithTeam:
        QUERY = gql(
            """
            mutation ProjectInviteCancel($projectId: ID!, $inviteId: String!) {
              data:projectMutations {
                data:invites {
                  data:cancel(projectId: $projectId, inviteId: $inviteId) {
                    id
                    name
                    description
                    visibility
                    allowPublicComments
                    role
                    createdAt
                    updatedAt
                    sourceApps
                    workspaceId
                    team {
                      id
                      role
                      user {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                      }
                    }
                    invitedTeam {
                      id
                      inviteId
                      projectId
                      projectName
                      title
                      role
                      token
                      user {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                      }
                      invitedBy {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                      }
                    }
                  }
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "inviteId": invite_id,
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[DataResponse[ProjectWithTeam]]], QUERY, variables
        ).data.data.data



from datetime import datetime
from typing import List, Optional, Union

from deprecated import deprecated

from specklepy.core.api.models import (
    ActivityCollection,
    LimitedUser,
    UserSearchResultCollection,
)
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources import OtherUserResource as CoreResource
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException


class OtherUserResource(CoreResource):
    """
    Provides API access to other users' profiles and activities on the platform.
    This class enables fetching limited information about users, searching for users by name or email,
    and accessing user activity logs with appropriate privacy and access control measures in place.
    """

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=(server_version,),
        )
        self.schema = LimitedUser

    def get(self, id: str) -> Optional[LimitedUser]:
        metrics.track(metrics.SDK, self.account, {"name": "Other User Get"})
        return super().get(id)

    def user_search(
        self,
        query: str,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        archived: bool = False,
        emailOnly: bool = False,
    ) -> UserSearchResultCollection:
        metrics.track(metrics.SDK, self.account, {"name": "Other User Search"})
        return super().user_search(
            query, limit=limit, cursor=cursor, archived=archived, emailOnly=emailOnly
        )

    @deprecated(reason="Use user_search instead", version=FE1_DEPRECATION_VERSION)
    def search(
        self, search_query: str, limit: int = 25
    ) -> Union[List[LimitedUser], SpeckleException]:
        """
        Searches for users by name or email.
        The search requires a minimum query length of 3 characters.

        Args:
            search_query (str): The search string.
            limit (int): Maximum number of search results to return.

        Returns:
            Union[List[LimitedUser], SpeckleException]: A list of users matching the search
                query or an exception if the query is too short.
        """
        if len(search_query) < 3:
            return SpeckleException(
                message="User search query must be at least 3 characters."
            )

        metrics.track(metrics.SDK, self.account, {"name": "Other User Search"})
        return super().search(search_query, limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def activity(
        self,
        user_id: str,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ) -> ActivityCollection:
        """
        Retrieves a collection of activities for a specified user, with optional filters for activity type,
        time frame, and pagination.

        Args:
            user_id (str): The ID of the user whose activities are being requested.
            limit (int): The maximum number of activity items to return.
            action_type (Optional[str]): A specific type of activity to filter.
            before (Optional[datetime]): Latest timestamp to include activities before.
            after (Optional[datetime]): Earliest timestamp to include activities after.
            cursor (Optional[datetime]): Timestamp for pagination cursor.

        Returns:
            ActivityCollection: A collection of user activities filtered according to specified criteria.
        """
        metrics.track(metrics.SDK, self.account, {"name": "Other User Activity"})
        return super().activity(user_id, limit, action_type, before, after, cursor)



# following imports seem to be unnecessary, but they need to stay
# to not break the scripts using these functions as non-core
from specklepy.core.api.models import (
    Activity,
    ActivityCollection,
    Branch,
    Branches,
    Collaborator,
    Commit,
    Commits,
    LimitedUser,
    Object,
    PendingStreamCollaborator,
    ServerInfo,
    Stream,
    Streams,
    User,
)

__all__ = [
    "Activity",
    "ActivityCollection",
    "Branch",
    "Branches",
    "Collaborator",
    "Commit",
    "Commits",
    "LimitedUser",
    "Object",
    "PendingStreamCollaborator",
    "ServerInfo",
    "Stream",
    "Streams",
    "User",
]



from specklepy.core.api.host_applications import (
    ARCGIS,
    ARCHICAD,
    AUTOCAD,
    BLENDER,
    CIVIL,
    CSIBRIDGE,
    DXF,
    DYNAMO,
    ETABS,
    EXCEL,
    GRASSHOPPER,
    GSA,
    MICROSTATION,
    NET,
    OPENBUILDINGS,
    OPENRAIL,
    OPENROADS,
    OTHER,
    POWERBI,
    PYTHON,
    QGIS,
    REVIT,
    RHINO,
    SAFE,
    SAP2000,
    SKETCHUP,
    TEKLASTRUCTURES,
    TOPSOLID,
    UNITY,
    UNREAL,
    HostApplication,
    HostAppVersion,
    _app_name_host_app_mapping,
    get_host_app_from_string,
)

# re-exporting stuff from the moved api module
__all__ = [
    "ARCGIS",
    "ARCHICAD",
    "AUTOCAD",
    "BLENDER",
    "CIVIL",
    "CSIBRIDGE",
    "DXF",
    "DYNAMO",
    "ETABS",
    "EXCEL",
    "GRASSHOPPER",
    "GSA",
    "MICROSTATION",
    "NET",
    "OPENBUILDINGS",
    "OPENRAIL",
    "OPENROADS",
    "OTHER",
    "POWERBI",
    "PYTHON",
    "QGIS",
    "REVIT",
    "RHINO",
    "SAFE",
    "SAP2000",
    "SKETCHUP",
    "TEKLASTRUCTURES",
    "TOPSOLID",
    "UNITY",
    "UNREAL",
    "HostApplication",
    "HostAppVersion",
    "_app_name_host_app_mapping",
    "get_host_app_from_string",
]



from deprecated import deprecated

from specklepy.core.api.models.deprecated import FE1_DEPRECATION_VERSION
from specklepy.core.api.resources import ServerResource

NAME = "server"


@deprecated(reason="Renamed to ServerResource", version=FE1_DEPRECATION_VERSION)
class Resource(ServerResource):
    """API Access class for the server"""



from typing import Dict, List

from gql import gql

from specklepy.core.api.resource import ResourceBase
from specklepy.objects.base import Base

NAME = "object"


class Resource(ResourceBase):
    """API Access class for objects"""

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
        )
        self.schema = Base

    def get(self, stream_id: str, object_id: str) -> Base:
        """
        Get a stream object

        Arguments:
            stream_id {str} -- the id of the stream for the object
            object_id {str} -- the hash of the object you want to get

        Returns:
            Base -- the returned Base object
        """
        query = gql(
            """
            query Object($stream_id: String!, $object_id: String!) {
              stream(id: $stream_id) {
                id
                name
                object(id: $object_id) {
                  id
                  speckleType
                  applicationId
                  createdAt
                  totalChildrenCount
                  data
                }
              }
            }
          """
        )
        params = {"stream_id": stream_id, "object_id": object_id}

        return self.make_request(
            query=query,
            params=params,
            return_type=["stream", "object", "data"],
        )

    def create(self, stream_id: str, objects: List[Dict]) -> str:
        """
        Not advised - generally, you want to use `operations.send()`.

        Create a new object on a stream.
        To send a base object, you can prepare it by running it through the
        `BaseObjectSerializer.traverse_base()` function to get a valid (serialisable)
        object to send.

        NOTE: this does not create a commit - you can create one with
        `SpeckleClient.commit.create`.
        Dynamic fields will be located in the 'data' dict of the received `Base` object

        Arguments:
            stream_id {str} -- the id of the stream you want to send the object to
            objects {List[Dict]}
            -- a list of base dictionary objects (NOTE: must be json serialisable)

        Returns:
            str -- the id of the object
        """
        query = gql(
            """
          mutation ObjectCreate($object_input: ObjectCreateInput!) {
            objectCreate(objectInput: $object_input)
        }
          """
        )
        params = {"object_input": {"streamId": stream_id, "objects": objects}}

        return self.make_request(
            query=query, params=params, return_type="objectCreate", parse_response=False
        )



from deprecated import deprecated

from specklepy.api.resources import ActiveUserResource
from specklepy.core.api.models.deprecated import FE1_DEPRECATION_VERSION


@deprecated(reason="Renamed to ActiveUserResource", version=FE1_DEPRECATION_VERSION)
class Resource(ActiveUserResource):
    """Renamed to ActiveUserResource"""



from datetime import datetime, timezone
from typing import List, Optional, Union

from deprecated import deprecated
from gql import gql

from specklepy.core.api.models import (
    ActivityCollection,
    PendingStreamCollaborator,
    User,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.logging.exceptions import SpeckleException

NAME = "user"

DEPRECATION_VERSION = "2.9.0"
DEPRECATION_TEXT = (
    "The user resource is deprecated, please use the active_user or other_user"
    " resources"
)


class Resource(ResourceBase):
    """API Access class for users"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )
        self.schema = User

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get(self, id: Optional[str] = None) -> User:
        """
        Gets the profile of a user.
        If no id argument is provided, will return the current authenticated
        user's profile (as extracted from the authorization header).

        Arguments:
            id {str} -- the user id

        Returns:
            User -- the retrieved user
        """
        query = gql(
            """
            query User($id: String) {
                user(id: $id) {
                    id
                    email
                    name
                    bio
                    company
                    avatar
                    verified
                    profiles
                    role
                }
            }
          """
        )

        params = {"id": id}

        return self.make_request(query=query, params=params, return_type="user")

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def search(
        self, search_query: str, limit: int = 25
    ) -> Union[List[User], SpeckleException]:
        """
        Searches for user by name or email.
        The search query must be at least 3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
        Returns:
            List[User] -- a list of User objects that match the search query
        """
        if len(search_query) < 3:
            return SpeckleException(
                message="User search query must be at least 3 characters"
            )

        query = gql(
            """
            query UserSearch($search_query: String!, $limit: Int!) {
                userSearch(query: $search_query, limit: $limit) {
                    items {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                    }
                }
            }
          """
        )
        params = {"search_query": search_query, "limit": limit}

        return self.make_request(
            query=query, params=params, return_type=["userSearch", "items"]
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
    ):
        """Updates your user profile. All arguments are optional.

        Arguments:
            name {str} -- your name
            company {str} -- the company you may or may not work for
            bio {str} -- tell us about yourself
            avatar {str} -- a nice photo of yourself

        Returns:
            bool -- True if your profile was updated successfully
        """
        query = gql(
            """
            mutation UserUpdate($user: UserUpdateInput!) {
                userUpdate(user: $user)
            }
            """
        )
        params = {"name": name, "company": company, "bio": bio, "avatar": avatar}

        params = {"user": {k: v for k, v in params.items() if v is not None}}

        if not params["user"]:
            return SpeckleException(
                message=(
                    "You must provide at least one field to update your user profile"
                )
            )

        return self.make_request(
            query=query, params=params, return_type="userUpdate", parse_response=False
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def activity(
        self,
        user_id: Optional[str] = None,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.
        If no id argument is provided, will return the current authenticated
        user's activity (as extracted from the authorization header).

        Note: all timestamps arguments should be `datetime` of any tz as
        they will be converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
        (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """

        query = gql(
            """
            query UserActivity(
                $user_id: String,
                $action_type: String,
                $before:DateTime,
                $after: DateTime,
                $cursor: DateTime,
                $limit: Int
                ){
                user(id: $user_id) {
                    activity(
                        actionType: $action_type,
                        before: $before,
                        after: $after,
                        cursor: $cursor,
                        limit: $limit
                        ) {
                        totalCount
                        cursor
                        items {
                            actionType
                            info
                            userId
                            streamId
                            resourceId
                            resourceType
                            message
                            time
                        }
                    }
                }
            }
            """
        )

        params = {
            "user_id": user_id,
            "limit": limit,
            "action_type": action_type,
            "before": before.astimezone(timezone.utc).isoformat() if before else before,
            "after": after.astimezone(timezone.utc).isoformat() if after else after,
            "cursor": cursor.astimezone(timezone.utc).isoformat() if cursor else cursor,
        }

        return self.make_request(
            query=query,
            params=params,
            return_type=["user", "activity"],
            schema=ActivityCollection,
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get_all_pending_invites(self) -> List[PendingStreamCollaborator]:
        """Get all of the active user's pending stream invites

        Requires Speckle Server version >= 2.6.4

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the current user
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvites {
                streamInvites{
                    id
                    token
                    inviteId
                    streamId
                    streamName
                    title
                    role
                    invitedBy {
                        id
                        name
                        company
                        avatar
                    }
                }
            }
            """
        )

        return self.make_request(
            query=query,
            return_type="streamInvites",
            schema=PendingStreamCollaborator,
        )

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get_pending_invite(
        self, stream_id: str, token: Optional[str] = None
    ) -> Optional[PendingStreamCollaborator]:
        """Get a particular pending invite for the active user on a given stream.
        If no invite_id is provided, any valid invite will be returned.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to look for invites on
            token {str} -- the token of the invite to look for (optional)

        Returns:
            PendingStreamCollaborator
                -- the invite for the given stream (or None if it isn't found)
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvite($streamId: String!, $token: String) {
                streamInvite(streamId: $streamId, token: $token) {
                    id
                    token
                    streamId
                    streamName
                    title
                    role
                    invitedBy {
                        id
                        name
                        company
                        avatar
                    }
                }
            }
            """
        )

        params = {"streamId": stream_id}
        if token:
            params["token"] = token

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInvite",
            schema=PendingStreamCollaborator,
        )



from datetime import datetime
from typing import List, Optional, overload

from deprecated import deprecated

from specklepy.core.api.inputs.project_inputs import UserProjectsFilter
from specklepy.core.api.inputs.user_inputs import UserUpdateInput
from specklepy.core.api.models import (
    PendingStreamCollaborator,
    Project,
    ResourceCollection,
    User,
)
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources import ActiveUserResource as CoreResource
from specklepy.logging import metrics


class ActiveUserResource(CoreResource):
    """API Access class for users. This class provides methods to get and update
    the user profile, fetch user activity, and manage pending stream invitations."""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )
        self.schema = User

    def get(self) -> Optional[User]:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Get"})
        return super().get()

    @deprecated("Use UserUpdateInput overload", version=FE1_DEPRECATION_VERSION)
    @overload
    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
    ) -> User:
        ...

    @overload
    def update(self, *, input: UserUpdateInput) -> User:
        ...

    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
        *,
        input: Optional[UserUpdateInput] = None,
    ) -> User:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Update"})
        if isinstance(input, UserUpdateInput):
            return super()._update(input=input)
        else:
            return super()._update(
                input=UserUpdateInput(
                    name=name,
                    company=company,
                    bio=bio,
                    avatar=avatar,
                )
            )

    def get_projects(
        self,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserProjectsFilter] = None,
    ) -> ResourceCollection[Project]:
        metrics.track(metrics.SDK, self.account, {"name": "Active User Get Projects"})
        return super().get_projects(limit=limit, cursor=cursor, filter=filter)

    def get_project_invites(self) -> List[PendingStreamCollaborator]:
        metrics.track(
            metrics.SDK, self.account, {"name": "Active User Get Project Invites"}
        )
        return super().get_project_invites()

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def activity(
        self,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Fetches collection the current authenticated user's activity
        as filtered by given parameters

        Note: all timestamps arguments should be `datetime` of any tz as they will be
        converted to UTC ISO format strings

        Args:
            limit (int): The maximum number of activity items to return.
            action_type (Optional[str]): Filter results to a single action type.
            before (Optional[datetime]): Latest cutoff for activity to include.
            after (Optional[datetime]): Oldest cutoff for an activity to include.
            cursor (Optional[datetime]): Timestamp cursor for pagination.

        Returns:
            Activity collection, filtered according to the provided parameters.
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Active Activity"})
        return super().activity(limit, action_type, before, after, cursor)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get_all_pending_invites(self) -> List[PendingStreamCollaborator]:
        """Fetches all of the current user's pending stream invitations.

        Returns:
            List[PendingStreamCollaborator]: A list of pending stream invitations.
        """
        metrics.track(
            metrics.SDK, self.account, {"name": "User Active Invites All Get"}
        )
        return super().get_all_pending_invites()

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get_pending_invite(
        self, stream_id: str, token: Optional[str] = None
    ) -> Optional[PendingStreamCollaborator]:
        """Fetches a specific pending invite for the current user on a given stream.

        Args:
            stream_id (str): The ID of the stream to look for invites on.
            token (Optional[str]): The token of the invite to look for (optional).

        Returns:
            Optional[PendingStreamCollaborator]: The invite for the given stream, or None if not found.
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Active Invite Get"})
        return super().get_pending_invite(stream_id, token)



import re
from typing import Dict
from warnings import warn

from deprecated import deprecated
from gql import Client
from gql.transport.exceptions import TransportServerError
from gql.transport.requests import RequestsHTTPTransport
from gql.transport.websockets import WebsocketsTransport

from specklepy.core.api import resources
from specklepy.core.api.credentials import Account, get_account_from_token
from specklepy.core.api.resources import (
    ActiveUserResource,
    ModelResource,
    OtherUserResource,
    ProjectInviteResource,
    ProjectResource,
    ServerResource,
    SubscriptionResource,
    VersionResource,
    branch,
    commit,
    object,
    stream,
    subscriptions,
    user,
)
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException, SpeckleWarning


class SpeckleClient:
    """
    The `SpeckleClient` is your entry point for interacting with
    your Speckle Server's GraphQL API.
    You'll need to have access to a server to use it,
    or you can use our public server `app.speckle.systems`.

    To authenticate the client, you'll need to have downloaded
    the [Speckle Manager](https://speckle.guide/#speckle-manager)
    and added your account.

    ```py
    from specklepy.api.client import SpeckleClient
    from specklepy.api.credentials import get_default_account

    # initialise the client
    client = SpeckleClient(host="app.speckle.systems") # or whatever your host is
    # client = SpeckleClient(host="localhost:3000", use_ssl=False) or use local server

    # authenticate the client with an account (account has been added in Speckle Manager)
    account = get_default_account()
    client.authenticate_with_account(account)

    # create a new stream. this returns the stream id
    new_stream_id = client.stream.create(name="a shiny new stream")

    # use that stream id to get the stream from the server
    new_stream = client.stream.get(id=new_stream_id)
    ```
    """

    DEFAULT_HOST = "app.speckle.systems"
    USE_SSL = True

    def __init__(
        self,
        host: str = DEFAULT_HOST,
        use_ssl: bool = USE_SSL,
        verify_certificate: bool = True,
        connection_retries: int = 3,
        connection_timeout: int = 10,
    ) -> None:
        ws_protocol = "ws"
        http_protocol = "http"

        if use_ssl:
            ws_protocol = "wss"
            http_protocol = "https"

        # sanitise host input by removing protocol and trailing slash
        host = re.sub(r"((^\w+:|^)\/\/)|(\/$)", "", host)

        self.url = f"{http_protocol}://{host}"
        self.graphql = f"{self.url}/graphql"
        self.ws_url = f"{ws_protocol}://{host}/graphql"
        self.account = Account()
        self.verify_certificate = verify_certificate
        self.connection_retries = connection_retries
        self.connection_timeout = connection_timeout

        self.httpclient = Client(
            transport=RequestsHTTPTransport(
                url=self.graphql,
                verify=self.verify_certificate,
                retries=self.connection_retries,
                timeout=self.connection_timeout,
            )
        )
        self.wsclient = None

        self._init_resources()

        # ? Check compatibility with the server - i think we can skip this at this point? save a request
        # try:
        #     server_info = self.server.get()
        #     if isinstance(server_info, Exception):
        #         raise server_info
        #     if not isinstance(server_info, ServerInfo):
        #         raise Exception("Couldn't get ServerInfo")
        # except Exception as ex:
        #     raise SpeckleException(
        #         f"{self.url} is not a compatible Speckle Server", ex
        #     ) from ex

    def __repr__(self):
        return (
            f"SpeckleClient( server: {self.url}, authenticated:"
            f" {self.account.token is not None} )"
        )

    @deprecated(
        version="2.6.0",
        reason=(
            "Renamed: please use `authenticate_with_account` or"
            " `authenticate_with_token` instead."
        ),
    )
    def authenticate(self, token: str) -> None:
        """Authenticate the client using a personal access token
        The token is saved in the client object and a synchronous GraphQL
        entrypoint is created

        Arguments:
            token {str} -- an api token
        """
        self.authenticate_with_account(get_account_from_token(token))

    def authenticate_with_token(self, token: str) -> None:
        """
        Authenticate the client using a personal access token.
        The token is saved in the client object and a synchronous GraphQL
        entrypoint is created

        Arguments:
            token {str} -- an api token
        """
        self.account = Account.from_token(token, self.url)
        self._set_up_client()

    def authenticate_with_account(self, account: Account) -> None:
        """Authenticate the client using an Account object
        The account is saved in the client object and a synchronous GraphQL
        entrypoint is created

        Arguments:
            account {Account} -- the account object which can be found with
            `get_default_account` or `get_local_accounts`
        """
        self.account = account
        self._set_up_client()

    def _set_up_client(self) -> None:
        headers = {
            "Authorization": f"Bearer {self.account.token}",
            "Content-Type": "application/json",
            "apollographql-client-name": metrics.HOST_APP,
            "apollographql-client-version": metrics.HOST_APP_VERSION,
        }
        httptransport = RequestsHTTPTransport(
            url=self.graphql, headers=headers, verify=self.verify_certificate, retries=3
        )
        wstransport = WebsocketsTransport(
            url=self.ws_url,
            init_payload={"Authorization": f"Bearer {self.account.token}"},
        )
        self.httpclient = Client(transport=httptransport)
        self.wsclient = Client(transport=wstransport)

        self._init_resources()

        try:
            _ = self.active_user.get()
        except SpeckleException as ex:
            if isinstance(ex.exception, TransportServerError):
                if ex.exception.code == 403:
                    warn(
                        SpeckleWarning(
                            "Possibly invalid token - could not authenticate Speckle Client"
                            f" for server {self.url}"
                        )
                    )
                else:
                    raise ex

    def execute_query(self, query: str) -> Dict:
        return self.httpclient.execute(query)

    def _init_resources(self) -> None:
        self.server = ServerResource(
            account=self.account, basepath=self.url, client=self.httpclient
        )

        server_version = None
        try:
            server_version = self.server.version()
        except Exception:
            pass

        self.other_user = OtherUserResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.active_user = ActiveUserResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.project = ProjectResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.project_invite = ProjectInviteResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.model = ModelResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.version = VersionResource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.subscription = SubscriptionResource(
            account=self.account,
            basepath=self.ws_url,
            client=self.wsclient,
        )
        # Deprecated Resources
        self.user = user.Resource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.stream = stream.Resource(
            account=self.account,
            basepath=self.url,
            client=self.httpclient,
            server_version=server_version,
        )
        self.commit = commit.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )
        self.branch = branch.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )
        self.object = object.Resource(
            account=self.account, basepath=self.url, client=self.httpclient
        )
        self.subscribe = subscriptions.Resource(
            account=self.account,
            basepath=self.ws_url,
            client=self.wsclient,
        )

    def __getattr__(self, name):
        try:
            attr = getattr(resources, name)
            return attr.Resource(
                account=self.account, basepath=self.url, client=self.httpclient
            )
        except AttributeError:
            raise SpeckleException(
                f"Method {name} is not supported by the SpeckleClient class"
            )



from datetime import datetime
from typing import List, Optional, Union

from deprecated import deprecated

from specklepy.api.models import PendingStreamCollaborator, User
from specklepy.core.api.resources.deprecated.user import Resource as CoreResource
from specklepy.logging import metrics
from specklepy.logging.exceptions import SpeckleException

DEPRECATION_VERSION = "2.9.0"
DEPRECATION_TEXT = (
    "The user resource is deprecated, please use the active_user or other_user"
    " resources"
)


class Resource(CoreResource):
    """API Access class for users"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )
        self.schema = User

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get(self, id: Optional[str] = None) -> User:
        """
        Gets the profile of a user.
        If no id argument is provided, will return the current authenticated
        user's profile (as extracted from the authorization header).

        Arguments:
            id {str} -- the user id

        Returns:
            User -- the retrieved user
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Get_deprecated"})
        return super().get(id)

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def search(
        self, search_query: str, limit: int = 25
    ) -> Union[List[User], SpeckleException]:
        """
        Searches for user by name or email.
        The search query must be at least 3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
        Returns:
            List[User] -- a list of User objects that match the search query
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Search_deprecated"})
        return super().search(search_query, limit)

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
    ):
        """Updates your user profile. All arguments are optional.

        Arguments:
            name {str} -- your name
            company {str} -- the company you may or may not work for
            bio {str} -- tell us about yourself
            avatar {str} -- a nice photo of yourself

        Returns:
            bool -- True if your profile was updated successfully
        """
        # metrics.track(metrics.USER, self.account, {"name": "update"})
        metrics.track(metrics.SDK, self.account, {"name": "User Update_deprecated"})
        return super().update(name, company, bio, avatar)

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def activity(
        self,
        user_id: Optional[str] = None,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.
        If no id argument is provided, will return the current authenticated
        user's activity (as extracted from the authorization header).

        Note: all timestamps arguments should be `datetime` of any tz as
        they will be converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
        (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        metrics.track(metrics.SDK, self.account, {"name": "User Activity_deprecated"})
        return super().activity(user_id, limit, action_type, before, after, cursor)

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get_all_pending_invites(self) -> List[PendingStreamCollaborator]:
        """Get all of the active user's pending stream invites

        Requires Speckle Server version >= 2.6.4

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the current user
        """
        # metrics.track(metrics.INVITE, self.account, {"name": "get"})
        metrics.track(
            metrics.SDK, self.account, {"name": "User GetAllInvites_deprecated"}
        )
        return super().get_all_pending_invites()

    @deprecated(version=DEPRECATION_VERSION, reason=DEPRECATION_TEXT)
    def get_pending_invite(
        self, stream_id: str, token: Optional[str] = None
    ) -> Optional[PendingStreamCollaborator]:
        """Get a particular pending invite for the active user on a given stream.
        If no invite_id is provided, any valid invite will be returned.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to look for invites on
            token {str} -- the token of the invite to look for (optional)

        Returns:
            PendingStreamCollaborator
                -- the invite for the given stream (or None if it isn't found)
        """
        # metrics.track(metrics.INVITE, self.account, {"name": "get"})
        metrics.track(metrics.SDK, self.account, {"name": "User GetInvite_deprecated"})
        return super().get_pending_invite(stream_id, token)



from typing import Optional

from gql import gql

from specklepy.core.api.inputs.model_inputs import ModelVersionsFilter
from specklepy.core.api.inputs.version_inputs import (
    CreateVersionInput,
    DeleteVersionsInput,
    MarkReceivedVersionInput,
    MoveVersionsInput,
    UpdateVersionInput,
)
from specklepy.core.api.models import ResourceCollection, Version
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse

NAME = "model"


class VersionResource(ResourceBase):
    """API Access class for model versions"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )

    def get(self, version_id: str, project_id: str) -> Version:
        QUERY = gql(
            """
            query VersionGet($projectId: String!, $versionId: String!) {
              data:project(id: $projectId) {
                data:version(id: $versionId) {
                  id
                  referencedObject
                  message
                  sourceApplication
                  createdAt
                  previewUrl
                  authorUser {
                    id
                    name
                    bio
                    company
                    verified
                    role
                    avatar
                  }
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "versionId": version_id,
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Version]], QUERY, variables
        ).data.data

    def get_versions(
        self,
        model_id: str,
        project_id: str,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[ModelVersionsFilter] = None,
    ) -> ResourceCollection[Version]:
        QUERY = gql(
            """
            query VersionGetVersions($projectId: String!, $modelId: String!, $limit: Int!, $cursor: String, $filter: ModelVersionsFilter) {
              data:project(id: $projectId) {
                data:model(id: $modelId) {
                  data:versions(limit: $limit, cursor: $cursor, filter: $filter) {
                    items {
                      id
                      referencedObject
                      message
                      sourceApplication
                      createdAt
                      previewUrl
                      authorUser {
                        id
                        name
                        bio
                        company
                        verified
                        role
                        avatar
                      }
                    }
                    cursor
                    totalCount
                  }
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "modelId": model_id,
            "limit": limit,
            "cursor": cursor,
            "filter": filter.model_dump(warnings="error") if filter else None,
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[DataResponse[ResourceCollection[Version]]]],
            QUERY,
            variables,
        ).data.data.data

    def create(self, input: CreateVersionInput) -> str:
        QUERY = gql(
            """
            mutation Create($input: CreateVersionInput!) {
              data:versionMutations {
                data:create(input: $input) {
                  data:id
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[DataResponse[str]]], QUERY, variables
        ).data.data.data

    def update(self, input: UpdateVersionInput) -> Version:
        QUERY = gql(
            """
            mutation VersionUpdate($input: UpdateVersionInput!) {
              data:versionMutations {
                data:update(input: $input) {
                  id
                  referencedObject
                  message
                  sourceApplication
                  createdAt
                  previewUrl
                  authorUser {
                    id
                    name
                    bio
                    company
                    verified
                    role
                    avatar
                  }
                }
              }
            }
            """
        )

        variables = {"input": input.model_dump(warnings="error")}

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Version]], QUERY, variables
        ).data.data

    def move_to_model(self, input: MoveVersionsInput) -> str:
        QUERY = gql(
            """
            mutation VersionMoveToModel($input: MoveVersionsInput!) {
              data:versionMutations {
                data:moveToModel(input: $input) {
                  data:id
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[DataResponse[str]]], QUERY, variables
        ).data.data.data

    def delete(self, input: DeleteVersionsInput) -> bool:
        QUERY = gql(
            """
            mutation VersionDelete($input: DeleteVersionsInput!) {
              data:versionMutations {
                data:delete(input: $input)
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[bool]], QUERY, variables
        ).data.data

    def received(self, input: MarkReceivedVersionInput) -> bool:
        QUERY = gql(
            """
            mutation MarkReceived($input: MarkReceivedVersionInput!) {
              data:versionMutations {
                data:markReceived(input: $input)
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[bool]], QUERY, variables
        ).data.data



from typing import Optional, Sequence

from pydantic import BaseModel

from specklepy.core.api.enums import ProjectVisibility


class ProjectCreateInput(BaseModel):
    name: Optional[str]
    description: Optional[str]
    visibility: Optional[ProjectVisibility]


class ProjectInviteCreateInput(BaseModel):
    email: Optional[str]
    role: Optional[str]
    serverRole: Optional[str]
    userId: Optional[str]


class ProjectInviteUseInput(BaseModel):
    accept: bool
    projectId: str
    token: str


class ProjectModelsFilter(BaseModel):
    contributors: Optional[Sequence[str]] = None
    excludeIds: Optional[Sequence[str]] = None
    ids: Optional[Sequence[str]] = None
    onlyWithVersions: Optional[bool] = None
    search: Optional[str] = None
    sourceApps: Optional[Sequence[str]] = None


class ProjectUpdateInput(BaseModel):
    id: str
    name: Optional[str] = None
    description: Optional[str] = None
    allowPublicComments: Optional[bool] = None
    visibility: Optional[ProjectVisibility] = None


class ProjectUpdateRoleInput(BaseModel):
    userId: str
    projectId: str
    role: Optional[str]


class UserProjectsFilter(BaseModel):
    search: str
    onlyWithRoles: Optional[Sequence[str]] = None



from datetime import datetime
from typing import Generic, List, Optional, TypeVar

from pydantic import BaseModel

from specklepy.core.api.enums import ProjectVisibility
from specklepy.core.api.models.deprecated import Streams

T = TypeVar("T")


class User(BaseModel):
    id: str
    email: Optional[str] = None
    name: str
    bio: Optional[str] = None
    company: Optional[str] = None
    avatar: Optional[str] = None
    verified: Optional[bool] = None
    role: Optional[str] = None
    streams: Optional["Streams"] = None

    def __repr__(self):
        return (
            f"User( id: {self.id}, name: {self.name}, email: {self.email}, company:"
            f" {self.company} )"
        )

    def __str__(self) -> str:
        return self.__repr__()


class ResourceCollection(BaseModel, Generic[T]):
    totalCount: int
    items: List[T]
    cursor: Optional[str] = None


class ServerMigration(BaseModel):
    movedFrom: Optional[str]
    movedTo: Optional[str]


class AuthStrategy(BaseModel):
    color: Optional[str]
    icon: str
    id: str
    name: str
    url: str


class ServerConfiguration(BaseModel):
    blobSizeLimitBytes: int
    objectMultipartUploadSizeLimitBytes: int
    objectSizeLimitBytes: int


# Keeping this one all Optionals at the minute, because its used both as a deserialization model for GQL and Account Management
class ServerInfo(BaseModel):
    name: Optional[str] = None
    company: Optional[str] = None
    url: Optional[str] = None
    adminContact: Optional[str] = None
    description: Optional[str] = None
    canonicalUrl: Optional[str] = None
    roles: Optional[List[dict]] = None
    scopes: Optional[List[dict]] = None
    authStrategies: Optional[List[dict]] = None
    version: Optional[str] = None
    frontend2: Optional[bool] = None
    migration: Optional[ServerMigration] = None

    # TODO separate gql model from account management model


class LimitedUser(BaseModel):
    """Limited user type, for showing public info about a user to another user."""

    id: str
    name: str
    bio: Optional[str]
    company: Optional[str]
    avatar: Optional[str]
    verified: Optional[bool]
    role: Optional[str]


class PendingStreamCollaborator(BaseModel):
    id: str
    inviteId: str
    streamId: Optional[str] = None
    projectId: str
    streamName: Optional[str] = None
    projectName: str
    title: str
    role: str
    invitedBy: LimitedUser
    user: Optional[LimitedUser] = None
    token: Optional[str]

    def __repr__(self):
        return (
            f"PendingStreamCollaborator( inviteId: {self.inviteId}, streamId:"
            f" {self.streamId}, role: {self.role}, title: {self.title}, invitedBy:"
            f" {self.user.name if self.user else None})"
        )

    def __str__(self) -> str:
        return self.__repr__()


class ProjectCollaborator(BaseModel):
    id: str
    role: str
    user: LimitedUser


class Version(BaseModel):
    authorUser: Optional[LimitedUser]
    createdAt: datetime
    id: str
    message: Optional[str]
    previewUrl: str
    referencedObject: str
    sourceApplication: Optional[str]


class Model(BaseModel):
    author: LimitedUser
    createdAt: datetime
    description: Optional[str]
    displayName: str
    id: str
    name: str
    previewUrl: Optional[str]
    updatedAt: datetime


class ModelWithVersions(Model):
    versions: ResourceCollection[Version]


class Project(BaseModel):
    allowPublicComments: bool
    createdAt: datetime
    description: Optional[str]
    id: str
    name: str
    role: Optional[str]
    sourceApps: List[str]
    updatedAt: datetime
    visibility: ProjectVisibility
    workspaceId: Optional[str]


class ProjectWithModels(Project):
    models: ResourceCollection[Model]


class ProjectWithTeam(Project):
    invitedTeam: List[PendingStreamCollaborator]
    team: List[ProjectCollaborator]


class ProjectCommentCollection(ResourceCollection[T], Generic[T]):
    totalArchivedCount: int


class UserSearchResultCollection(BaseModel):
    items: List[LimitedUser]
    cursor: Optional[str] = None



from typing import Any, Optional, Tuple

from gql.client import Client

from specklepy.api.credentials import Account
from specklepy.core.api.resource import ResourceBase as CoreResourceBase


class ResourceBase(CoreResourceBase):
    def __init__(
        self,
        account: Account,
        basepath: str,
        client: Client,
        name: str,
        server_version: Optional[Tuple[Any, ...]] = None,
    ) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=name,
            server_version=server_version,
        )



from datetime import datetime, timezone
from typing import List, Optional, overload

from deprecated import deprecated
from gql import gql

from specklepy.core.api.inputs.project_inputs import UserProjectsFilter
from specklepy.core.api.inputs.user_inputs import UserUpdateInput
from specklepy.core.api.models import (
    ActivityCollection,
    PendingStreamCollaborator,
    Project,
    ResourceCollection,
    User,
)
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse
from specklepy.logging.exceptions import GraphQLException

NAME = "active_user"


class ActiveUserResource(ResourceBase):
    """API Access class for the active user"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )
        self.schema = User

    def get(self) -> Optional[User]:
        """Gets the currently active user profile (as extracted from the authorization header)

        Returns:
            User -- the requested user, or none if no authentication token is provided to the Client
        """
        QUERY = gql(
            """
            query User {
             data:activeUser {
               id
               email
               name
               bio
               company
               avatar
               verified
               role
             }
           }
           """
        )

        variables = {}

        return self.make_request_and_parse_response(
            DataResponse[Optional[User]], QUERY, variables
        ).data

    def _update(self, input: UserUpdateInput) -> User:
        QUERY = gql(
            """
            mutation ActiveUserMutations($input: UserUpdateInput!) {
              data:activeUserMutations {
                data:update(user: $input) {
                  id
                  email
                  name
                  bio
                  company
                  avatar
                  verified
                  role
                }
              }
            }
           """
        )

        variables = {"input": input.model_dump(warnings="error")}

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[User]], QUERY, variables
        ).data.data

    @deprecated("Use UserUpdateInput overload", version=FE1_DEPRECATION_VERSION)
    @overload
    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
    ) -> User:
        ...

    @overload
    def update(self, *, input: UserUpdateInput) -> User:
        ...

    def update(
        self,
        name: Optional[str] = None,
        company: Optional[str] = None,
        bio: Optional[str] = None,
        avatar: Optional[str] = None,
        *,
        input: Optional[UserUpdateInput] = None,
    ) -> User:
        if isinstance(input, UserUpdateInput):
            return self._update(input=input)
        else:
            return self._update(
                input=UserUpdateInput(
                    name=name,
                    company=company,
                    bio=bio,
                    avatar=avatar,
                )
            )

    def get_projects(
        self,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[UserProjectsFilter] = None,
    ) -> ResourceCollection[Project]:
        QUERY = gql(
            """
             query User($limit : Int!, $cursor: String, $filter: UserProjectsFilter) {
              data:activeUser {
                data:projects(limit: $limit, cursor: $cursor, filter: $filter) {
                   totalCount
                   cursor
                   items {
                      id
                      name
                      description
                      visibility
                      allowPublicComments
                      role
                      createdAt
                      updatedAt
                      sourceApps
                      workspaceId
                   }
                }
              }
            }
            """
        )

        variables = {
            "limit": limit,
            "cursor": cursor,
            "filter": filter.model_dump(warnings="error") if filter else None,
        }

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[ResourceCollection[Project]]]],
            QUERY,
            variables,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data

    def get_project_invites(self) -> List[PendingStreamCollaborator]:
        QUERY = gql(
            """
            query ProjectInvites {
              data:activeUser {
                data:projectInvites {
                  id
                  inviteId
                  invitedBy {
                    avatar
                    bio
                    company
                    id
                    name
                    role
                    verified
                  }
                  projectId
                  projectName
                  role
                  title
                  token
                  user {
                    id
                    name
                    bio
                    company
                    verified
                    avatar
                    role
                  }
                }
              }
            }
            """
        )

        variables = {}

        response = self.make_request_and_parse_response(
            DataResponse[Optional[DataResponse[List[PendingStreamCollaborator]]]],
            QUERY,
            variables,
        )

        if response.data is None:
            raise GraphQLException(
                "GraphQL response indicated that the ActiveUser could not be found"
            )

        return response.data.data

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def activity(
        self,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ) -> ActivityCollection:
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.
        If no id argument is provided, will return the current authenticated user's
        activity (as extracted from the authorization header).

        Note: all timestamps arguments should be `datetime` of any tz as they will be
        converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime} -- latest cutoff for activity
            (ie: return all activity _before_ this time)
        after {datetime} -- oldest cutoff for activity
            (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """

        query = gql(
            """
            query UserActivity(
                $action_type: String,
                $before:DateTime,
                $after: DateTime,
                $cursor: DateTime,
                $limit: Int
                ){
                activeUser {
                    activity(
                        actionType: $action_type,
                        before: $before,
                        after: $after,
                        cursor: $cursor,
                        limit: $limit
                        ) {
                        totalCount
                        cursor
                        items {
                            actionType
                            info
                            userId
                            streamId
                            resourceId
                            resourceType
                            message
                            time
                        }
                    }
                }
            }
            """
        )

        params = {
            "limit": limit,
            "action_type": action_type,
            "before": before.astimezone(timezone.utc).isoformat() if before else before,
            "after": after.astimezone(timezone.utc).isoformat() if after else after,
            "cursor": cursor.astimezone(timezone.utc).isoformat() if cursor else cursor,
        }

        return self.make_request(
            query=query,
            params=params,
            return_type=["activeUser", "activity"],
            schema=ActivityCollection,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get_all_pending_invites(self) -> List[PendingStreamCollaborator]:
        """Get all of the active user's pending stream invites

        Requires Speckle Server version >= 2.6.4

        Returns:
            List[PendingStreamCollaborator]
            -- a list of pending invites for the current user
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvites {
                streamInvites{
                    id
                    token
                    inviteId
                    streamId
                    streamName
                    projectId
                    projectName
                    title
                    role
                    invitedBy {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                    }
                }
            }
            """
        )

        return self.make_request(
            query=query,
            return_type="streamInvites",
            schema=PendingStreamCollaborator,
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get_pending_invite(
        self, stream_id: str, token: Optional[str] = None
    ) -> Optional[PendingStreamCollaborator]:
        """Get a particular pending invite for the active user on a given stream.
        If no invite_id is provided, any valid invite will be returned.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to look for invites on
            token {str} -- the token of the invite to look for (optional)

        Returns:
            PendingStreamCollaborator
            -- the invite for the given stream (or None if it isn't found)
        """
        self._check_invites_supported()

        query = gql(
            """
            query StreamInvite($streamId: String!, $token: String) {
                streamInvite(streamId: $streamId, token: $token) {
                    id
                    token
                    inviteId
                    streamId
                    streamName
                    projectId
                    projectName
                    title
                    role
                    invitedBy {
                        id
                        name
                        bio
                        company
                        avatar
                        verified
                        role
                    }
                }
            }
            """
        )

        params = {"streamId": stream_id}
        if token:
            params["token"] = token

        return self.make_request(
            query=query,
            params=params,
            return_type="streamInvite",
            schema=PendingStreamCollaborator,
        )



from datetime import datetime, timezone
from typing import List, Optional, Union

from deprecated import deprecated
from gql import gql

from specklepy.core.api.models import (
    ActivityCollection,
    LimitedUser,
    UserSearchResultCollection,
)
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse
from specklepy.logging.exceptions import SpeckleException

NAME = "other_user"


class OtherUserResource(ResourceBase):
    """API Access class for other users, that are not the currently active user."""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )
        self.schema = LimitedUser

    def get(self, id: str) -> Optional[LimitedUser]:
        """
        Gets the profile of another user.

        Arguments:
            id {str} -- the user id

        Returns:
            LimitedUser -- the retrieved profile of another user
        """
        QUERY = gql(
            """
          query LimitedUser($id: String!) {
            data:otherUser(id: $id){
              id
              name
              bio
              company
              avatar
              verified
              role
            }
          }
          """
        )

        variables = {"id": id}

        return self.make_request_and_parse_response(
            DataResponse[Optional[LimitedUser]], QUERY, variables
        ).data

    def user_search(
        self,
        query: str,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        archived: bool = False,
        emailOnly: bool = False,
    ) -> UserSearchResultCollection:
        """Searches for a user on the server, by name or email. The search query must be at least
        3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
            cursor {Optional[str]} --
            archived {bool} --
            emailOnly {bool} --
        Returns:
            ResourceCollection[LimitedUser] -- User objects that match the search query
        """

        QUERY = gql(
            """
            query UserSearch($query: String!, $limit: Int!, $cursor: String, $archived: Boolean, $emailOnly: Boolean) {
              data:userSearch(query: $query, limit: $limit, cursor: $cursor, archived: $archived, emailOnly: $emailOnly) {
                cursor
                items {
                 id
                 name
                 bio
                 company
                 avatar
                 verified
                 role
               }
             }
            }
            """
        )
        variables = {
            "query": query,
            "limit": limit,
            "cursor": cursor,
            "archived": archived,
            "emailOnly": emailOnly,
        }

        return self.make_request_and_parse_response(
            DataResponse[UserSearchResultCollection], QUERY, variables
        ).data

    @deprecated(reason="Use user_search instead", version=FE1_DEPRECATION_VERSION)
    def search(
        self, search_query: str, limit: int = 25
    ) -> Union[List[LimitedUser], SpeckleException]:
        """Searches for user by name or email. The search query must be at least
        3 characters long

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
        Returns:
            List[LimitedUser] -- a list of User objects that match the search query
        """
        if len(search_query) < 3:
            return SpeckleException(
                message="User search query must be at least 3 characters"
            )

        query = gql(
            """
            query UserSearch($search_query: String!, $limit: Int!) {
                userSearch(query: $search_query, limit: $limit) {
                    items {
                      id
                      name
                      bio
                      company
                      avatar
                      verified
                      role
                    }
                }
            }
          """
        )
        params = {"search_query": search_query, "limit": limit}

        return self.make_request(
            query=query, params=params, return_type=["userSearch", "items"]
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def activity(
        self,
        user_id: str,
        limit: int = 20,
        action_type: Optional[str] = None,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ) -> ActivityCollection:
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.

        Note: all timestamps arguments should be `datetime` of
        any tz as they will be converted to UTC ISO format strings

        user_id {str} -- the id of the user to get the activity from
        action_type {str} -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime} -- latest cutoff for activity
            (ie: return all activity _before_ this time)
        after {datetime} -- oldest cutoff for activity
            (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """

        query = gql(
            """
            query UserActivity(
                $user_id: String!,
                $action_type: String,
                $before:DateTime,
                $after: DateTime,
                $cursor: DateTime,
                $limit: Int
                ){
                otherUser(id: $user_id) {
                    activity(
                        actionType: $action_type,
                        before: $before,
                        after: $after,
                        cursor: $cursor,
                        limit: $limit
                        ) {
                        totalCount
                        cursor
                        items {
                            actionType
                            info
                            userId
                            streamId
                            resourceId
                            resourceType
                            message
                            time
                        }
                    }
                }
            }
            """
        )

        params = {
            "user_id": user_id,
            "limit": limit,
            "action_type": action_type,
            "before": before.astimezone(timezone.utc).isoformat() if before else before,
            "after": after.astimezone(timezone.utc).isoformat() if after else after,
            "cursor": cursor.astimezone(timezone.utc).isoformat() if cursor else cursor,
        }

        return self.make_request(
            query=query,
            params=params,
            return_type=["otherUser", "activity"],
            schema=ActivityCollection,
        )



from typing import Optional

from pydantic import BaseModel

from specklepy.core.api.enums import (
    ProjectModelsUpdatedMessageType,
    ProjectUpdatedMessageType,
    ProjectVersionsUpdatedMessageType,
    UserProjectsUpdatedMessageType,
)
from specklepy.core.api.models.current import Model, Project, Version


class UserProjectsUpdatedMessage(BaseModel):
    id: str
    type: UserProjectsUpdatedMessageType
    project: Optional[Project]


class ProjectModelsUpdatedMessage(BaseModel):
    id: str
    type: ProjectModelsUpdatedMessageType
    model: Optional[Model]


class ProjectUpdatedMessage(BaseModel):
    id: str
    type: ProjectUpdatedMessageType
    project: Optional[Project]


class ProjectVersionsUpdatedMessage(BaseModel):
    id: str
    type: ProjectVersionsUpdatedMessageType
    modelId: Optional[str]
    version: Optional[Version]



from datetime import datetime
from typing import List, Optional

from deprecated import deprecated

from specklepy.api.models import PendingStreamCollaborator, Stream
from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
)
from specklepy.core.api.resources.deprecated.stream import Resource as CoreResource
from specklepy.logging import metrics


class Resource(CoreResource):
    """API Access class for streams"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )

        self.schema = Stream

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get(self, id: str, branch_limit: int = 10, commit_limit: int = 10) -> Stream:
        """Get the specified stream from the server

        Arguments:
            id {str} -- the stream id
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            Stream -- the retrieved stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Get"})
        return super().get(id, branch_limit, commit_limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def list(self, stream_limit: int = 10) -> List[Stream]:
        """Get a list of the user's streams

        Arguments:
            stream_limit {int} -- The maximum number of streams to return

        Returns:
            List[Stream] -- A list of Stream objects
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream List"})
        return super().list(stream_limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def create(
        self,
        name: str = "Anonymous Python Stream",
        description: str = "No description provided",
        is_public: bool = True,
    ) -> str:
        """Create a new stream

        Arguments:
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            id {str} -- the id of the newly created stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Create"})
        return super().create(name, description, is_public)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update(
        self,
        id: str,
        name: Optional[str] = None,
        description: Optional[str] = None,
        is_public: Optional[bool] = None,
    ) -> bool:
        """Update an existing stream

        Arguments:
            id {str} -- the id of the stream to be updated
            name {str} -- the name of the string
            description {str} -- a short description of the stream
            is_public {bool}
                -- whether or not the stream can be viewed by anyone with the id

        Returns:
            bool -- whether the stream update was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Update"})
        return super().update(id, name, description, is_public)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def delete(self, id: str) -> bool:
        """Delete a stream given its id

        Arguments:
            id {str} -- the id of the stream to delete

        Returns:
            bool -- whether the deletion was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Delete"})
        return super().delete(id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def search(
        self,
        search_query: str,
        limit: int = 25,
        branch_limit: int = 10,
        commit_limit: int = 10,
    ):
        """Search for streams by name, description, or id

        Arguments:
            search_query {str} -- a string to search for
            limit {int} -- the maximum number of results to return
            branch_limit {int} -- the maximum number of branches to return
            commit_limit {int} -- the maximum number of commits to return

        Returns:
            List[Stream] -- a list of Streams that match the search query
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Search"})
        return super().search(search_query, limit, branch_limit, commit_limit)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def favorite(self, stream_id: str, favorited: bool = True):
        """Favorite or unfavorite the given stream.

        Arguments:
            stream_id {str} -- the id of the stream to favorite / unfavorite
            favorited {bool}
                -- whether to favorite (True) or unfavorite (False) the stream

        Returns:
            Stream -- the stream with its `id`, `name`, and `favoritedDate`
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Favorite"})
        return super().favorite(stream_id, favorited)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get_all_pending_invites(
        self, stream_id: str
    ) -> List[PendingStreamCollaborator]:
        """Get all of the pending invites on a stream.
        You must be a `stream:owner` to query this.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the stream id from which to get the pending invites

        Returns:
            List[PendingStreamCollaborator]
                -- a list of pending invites for the specified stream
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Get"})
        return super().get_all_pending_invites(stream_id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite(
        self,
        stream_id: str,
        email: Optional[str] = None,
        user_id: Optional[str] = None,
        role: str = "stream:contributor",  # should default be reviewer?
        message: Optional[str] = None,
    ):
        """Invite someone to a stream using either their email or user id

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            email {str} -- the email of the user to invite (use this OR `user_id`)
            user_id {str} -- the id of the user to invite (use this OR `email`)
            role {str}
                -- the role to assign to the user (defaults to `stream:contributor`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Create"})
        return super().invite(stream_id, email, user_id, role, message)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite_batch(
        self,
        stream_id: str,
        emails: Optional[List[str]] = None,
        user_ids: Optional[List[None]] = None,
        message: Optional[str] = None,
    ) -> bool:
        """Invite a batch of users to a specified stream.

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to invite the user to
            emails {List[str]}
                -- the email of the user to invite (use this and/or `user_ids`)
            user_id {List[str]}
                -- the id of the user to invite (use this and/or `emails`)
            message {str}
                -- a message to send along with this invite to the specified user

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Batch Create"})
        return super().invite_batch(stream_id, emails, user_ids, message)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite_cancel(self, stream_id: str, invite_id: str) -> bool:
        """Cancel an existing stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str} -- the id of the stream invite
            invite_id {str} -- the id of the invite to use

        Returns:
            bool -- true if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Invite Cancel"})
        return super().invite_cancel(stream_id, invite_id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def invite_use(self, stream_id: str, token: str, accept: bool = True) -> bool:
        """Accept or decline a stream invite

        Requires Speckle Server version >= 2.6.4

        Arguments:
            stream_id {str}
                -- the id of the stream for which the user has a pending invite
            token {str} -- the token of the invite to use
            accept {bool} -- whether or not to accept the invite (defaults to True)

        Returns:
            bool -- true if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Invite Use"})
        return super().invite_use(stream_id, token, accept)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update_permission(self, stream_id: str, user_id: str, role: str):
        """Updates permissions for a user on a given stream

        Valid for Speckle Server >=2.6.4

        Arguments:
            stream_id {str} -- the id of the stream to grant permissions to
            user_id {str} -- the id of the user to grant permissions for
            role {str} -- the role to grant the user

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(
            metrics.SDK,
            self.account,
            {"name": "Stream Permission Update", "role": role},
        )
        return super().update_permission(stream_id, user_id, role)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def revoke_permission(self, stream_id: str, user_id: str):
        """Revoke permissions from a user on a given stream

        Arguments:
            stream_id {str} -- the id of the stream to revoke permissions from
            user_id {str} -- the id of the user to revoke permissions from

        Returns:
            bool -- True if the operation was successful
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Permission Revoke"})
        return super().revoke_permission(stream_id, user_id)

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def activity(
        self,
        stream_id: str,
        action_type: Optional[str] = None,
        limit: int = 20,
        before: Optional[datetime] = None,
        after: Optional[datetime] = None,
        cursor: Optional[datetime] = None,
    ):
        """
        Get the activity from a given stream in an Activity collection.
        Step into the activity `items` for the list of activity.

        Note: all timestamps arguments should be `datetime` of any tz
            as they will be converted to UTC ISO format strings

        stream_id {str} -- the id of the stream to get activity from
        action_type {str}
            -- filter results to a single action type
            (eg: `commit_create` or `commit_receive`)
        limit {int} -- max number of Activity items to return
        before {datetime}
            -- latest cutoff for activity (ie: return all activity _before_ this time)
        after {datetime}
            -- oldest cutoff for activity (ie: return all activity _after_ this time)
        cursor {datetime} -- timestamp cursor for pagination
        """
        metrics.track(metrics.SDK, self.account, {"name": "Stream Activity"})
        return super().activity(stream_id, action_type, limit, before, after, cursor)



from deprecated import deprecated

from specklepy.api.resources import OtherUserResource
from specklepy.core.api.models.deprecated import FE1_DEPRECATION_VERSION


@deprecated(reason="Renamed to OtherUserResource", version=FE1_DEPRECATION_VERSION)
class Resource(OtherUserResource):
    """
    Renamed to OtherUserResource
    """



from typing import List, Optional

from specklepy.objects.base import Base

STRUCTURAL_ANALYSIS = "Objects.Structural.Analysis."


class ModelUnits(Base, speckle_type=STRUCTURAL_ANALYSIS + "ModelUnits"):
    length: Optional[str] = None
    sections: Optional[str] = None
    displacements: Optional[str] = None
    stress: Optional[str] = None
    force: Optional[str] = None
    mass: Optional[str] = None
    time: Optional[str] = None
    temperature: Optional[str] = None
    velocity: Optional[str] = None
    acceleration: Optional[str] = None
    energy: Optional[str] = None
    angle: Optional[str] = None
    strain: Optional[str] = None


class ModelSettings(Base, speckle_type=STRUCTURAL_ANALYSIS + "ModelSettings"):
    modelUnits: Optional[ModelUnits] = None
    steelCode: Optional[str] = None
    concreteCode: Optional[str] = None
    coincidenceTolerance: float = 0.0


class ModelInfo(Base, speckle_type=STRUCTURAL_ANALYSIS + "ModelInfo"):
    name: Optional[str] = None
    description: Optional[str] = None
    projectNumber: Optional[str] = None
    projectName: Optional[str] = None
    settings: Optional[ModelSettings] = None
    initials: Optional[str] = None
    application: Optional[str] = None


class Model(Base, speckle_type=STRUCTURAL_ANALYSIS + "Model"):
    specs: Optional[ModelInfo] = None
    nodes: Optional[List] = None
    elements: Optional[List] = None
    loads: Optional[List] = None
    restraints: Optional[List] = None
    properties: Optional[List] = None
    materials: Optional[List] = None
    layerDescription: Optional[str] = None



from typing import Any, List, Optional


class SpeckleException(Exception):
    def __init__(self, message: str, exception: Exception = None) -> None:
        super().__init__()
        self.message = message
        self.exception = exception

    def __str__(self) -> str:
        return f"SpeckleException: {self.message}"


class SpeckleInvalidUnitException(SpeckleException):
    def __init__(self, invalid_unit: Any) -> None:
        super().__init__(
            message=(
                "Invalid units: expected type str but received"
                f" {type(invalid_unit)} ({invalid_unit})."
            ),
            exception=None,
        )


class SerializationException(SpeckleException):
    def __init__(self, message: str, obj: Any, exception: Exception = None) -> None:
        super().__init__(message=message, exception=exception)
        self.obj = obj
        self.unhandled_type = type(obj)

    def __str__(self) -> str:
        return (
            "SpeckleException: Could not serialize object of type"
            f" {self.unhandled_type}"
        )


class GraphQLException(SpeckleException):
    def __init__(
        self, message: str, errors: Optional[List[Any]] = None, data=None
    ) -> None:
        super().__init__(message=message)
        self.errors = errors
        self.data = data

    def __str__(self) -> str:
        return f"GraphQLException: {self.message}"


class UnsupportedException(SpeckleException):
    def __init__(self, message: str) -> None:
        super().__init__(message=message)

    def __str__(self) -> str:
        return f"UnsupportedException: {self.message}"


class SpeckleWarning(Warning):
    def __init__(self, *args: object) -> None:
        super().__init__(*args)



from typing import Optional

from gql import gql

from specklepy.core.api.inputs.project_inputs import (
    ProjectCreateInput,
    ProjectModelsFilter,
    ProjectUpdateInput,
    ProjectUpdateRoleInput,
)
from specklepy.core.api.models import Project, ProjectWithModels, ProjectWithTeam
from specklepy.core.api.resource import ResourceBase
from specklepy.core.api.responses import DataResponse

NAME = "project"


class ProjectResource(ResourceBase):
    """API Access class for projects"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
            server_version=server_version,
        )

    def get(self, project_id: str) -> Project:
        QUERY = gql(
            """
            query Project($projectId: String!) {
              data:project(id: $projectId) {
                allowPublicComments
                createdAt
                description
                id
                name
                role
                sourceApps
                updatedAt
                visibility
                workspaceId
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
        }

        return self.make_request_and_parse_response(
            DataResponse[Project], QUERY, variables
        ).data

    def get_with_models(
        self,
        project_id: str,
        *,
        models_limit: int = 25,
        models_cursor: Optional[str] = None,
        models_filter: Optional[ProjectModelsFilter] = None,
    ) -> ProjectWithModels:
        QUERY = gql(
            """
            query ProjectGetWithModels($projectId: String!, $modelsLimit: Int!, $modelsCursor: String, $modelsFilter: ProjectModelsFilter) {
              data:project(id: $projectId) {
                id
                name
                description
                visibility
                allowPublicComments
                role
                createdAt
                updatedAt
                sourceApps
                workspaceId
                models(limit: $modelsLimit, cursor: $modelsCursor, filter: $modelsFilter) {
                  items {
                    id
                    name
                    previewUrl
                    updatedAt
                    displayName
                    description
                    createdAt
                    author {
                      avatar
                      bio
                      company
                      id
                      name
                      role
                      verified
                    }
                  }
                  cursor
                  totalCount
                }
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
            "modelsLimit": models_limit,
            "modelsCursor": models_cursor,
            "modelsFilter": models_filter.model_dump(warnings="error")
            if models_filter
            else None,
        }

        return self.make_request_and_parse_response(
            DataResponse[ProjectWithModels], QUERY, variables
        ).data

    def get_with_team(self, project_id: str) -> ProjectWithTeam:
        QUERY = gql(
            """
            query ProjectGetWithTeam($projectId: String!) {
              data:project(id: $projectId) {
                id
                name
                description
                visibility
                allowPublicComments
                role
                createdAt
                updatedAt
                workspaceId
                sourceApps
                team {
                  id
                  role
                  user {
                    id
                    name
                    bio
                    company
                    avatar
                    verified
                    role
                  }
                }
                invitedTeam {
                  id
                  inviteId
                  projectId
                  projectName
                  title
                  role
                  token
                  user {
                    id
                    name
                    bio
                    company
                    avatar
                    verified
                    role
                  }
                  invitedBy {
                    id
                    name
                    bio
                    company
                    avatar
                    verified
                    role
                  }
                }
                workspaceId
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
        }

        return self.make_request_and_parse_response(
            DataResponse[ProjectWithTeam], QUERY, variables
        ).data

    def create(self, input: ProjectCreateInput) -> Project:
        QUERY = gql(
            """
            mutation ProjectCreate($input: ProjectCreateInput) {
              data:projectMutations {
                data:create(input: $input) {
                  id
                  name
                  description
                  visibility
                  allowPublicComments
                  role
                  createdAt
                  updatedAt
                  sourceApps
                  workspaceId
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Project]], QUERY, variables
        ).data.data

    def update(self, input: ProjectUpdateInput) -> Project:
        QUERY = gql(
            """
            mutation ProjectUpdate($input: ProjectUpdateInput!) {
              data:projectMutations{
                data:update(update: $input) {
                  allowPublicComments
                  createdAt
                  description
                  id
                  name
                  role
                  sourceApps
                  updatedAt
                  visibility
                  workspaceId
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[Project]], QUERY, variables
        ).data.data

    def delete(self, project_id: str) -> bool:
        QUERY = gql(
            """
            mutation ProjectDelete($projectId: String!) {
              data:projectMutations {
                data:delete(id: $projectId)
              }
            }
            """
        )

        variables = {
            "projectId": project_id,
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[bool]], QUERY, variables
        ).data.data

    def update_role(self, input: ProjectUpdateRoleInput) -> ProjectWithTeam:
        QUERY = gql(
            """
            mutation ProjectUpdateRole($input: ProjectUpdateRoleInput!) {
              data:projectMutations {
                data:updateRole(input: $input) {
                  id
                  name
                  description
                  visibility
                  allowPublicComments
                  role
                  createdAt
                  updatedAt
                  sourceApps
                  workspaceId
                  team {
                    id
                    role
                    user {
                      id
                      name
                      bio
                      company
                      avatar
                      verified
                      role
                    }
                  }
                  invitedTeam {
                    id
                    inviteId
                    projectId
                    projectName
                    title
                    role
                    token
                    user {
                      id
                      name
                      bio
                      company
                      avatar
                      verified
                      role
                    }
                    invitedBy {
                      id
                      name
                      bio
                      company
                      avatar
                      verified
                      role
                    }
                  }
                }
              }
            }
            """
        )

        variables = {
            "input": input.model_dump(warnings="error"),
        }

        return self.make_request_and_parse_response(
            DataResponse[DataResponse[ProjectWithTeam]], QUERY, variables
        ).data.data



from typing import List, Optional, Union

from deprecated import deprecated
from gql import gql

from specklepy.core.api.models.deprecated import (
    FE1_DEPRECATION_REASON,
    FE1_DEPRECATION_VERSION,
    Commit,
)
from specklepy.core.api.resource import ResourceBase
from specklepy.logging.exceptions import SpeckleException

NAME = "commit"


class Resource(ResourceBase):
    """
    API Access class for commits
    Commit resource is deprecated, please use version resource instead
    """

    def __init__(self, account, basepath, client) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            name=NAME,
        )
        self.schema = Commit

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def get(self, stream_id: str, commit_id: str) -> Commit:
        """
        Gets a commit given a stream and the commit id

        Arguments:
            stream_id {str} -- the stream where we can find the commit
            commit_id {str} -- the id of the commit you want to get

        Returns:
            Commit -- the retrieved commit object
        """
        query = gql(
            """
            query Commit($stream_id: String!, $commit_id: String!) {
                stream(id: $stream_id) {
                    commit(id: $commit_id) {
                        id
                        message
                        referencedObject
                        authorId
                        authorName
                        authorAvatar
                        branchName
                        createdAt
                        sourceApplication
                        totalChildrenCount
                        parents
                    }
                }
            }
            """
        )
        params = {"stream_id": stream_id, "commit_id": commit_id}

        return self.make_request(
            query=query, params=params, return_type=["stream", "commit"]
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def list(self, stream_id: str, limit: int = 10) -> List[Commit]:
        """
        Get a list of commits on a given stream

        Arguments:
            stream_id {str} -- the stream where the commits are
            limit {int} -- the maximum number of commits to fetch (default = 10)

        Returns:
            List[Commit] -- a list of the most recent commit objects
        """
        query = gql(
            """
            query Commits($stream_id: String!, $limit: Int!) {
                stream(id: $stream_id) {
                    commits(limit: $limit) {
                        items {
                            id
                            message
                            referencedObject
                            authorName
                            authorId
                            authorName
                            authorAvatar
                            branchName
                            createdAt
                            sourceApplication
                            totalChildrenCount
                            parents
                        }
                    }
                }
            }
            """
        )
        params = {"stream_id": stream_id, "limit": limit}

        return self.make_request(
            query=query, params=params, return_type=["stream", "commits", "items"]
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def create(
        self,
        stream_id: str,
        object_id: str,
        branch_name: str = "main",
        message: str = "",
        source_application: str = "python",
        parents: Optional[List[str]] = None,
    ) -> Union[str, SpeckleException]:
        """
        Creates a commit on a branch

        Arguments:
            stream_id {str} -- the stream you want to commit to
            object_id {str} -- the hash of your commit object
            branch_name {str}
                -- the name of the branch to commit to (defaults to "main")
            message {str}
                -- optional: a message to give more information about the commit
            source_application{str}
                -- optional: the application from which the commit was created
                (defaults to "python")
            parents {List[str]} -- optional: the id of the parent commits

        Returns:
            str -- the id of the created commit
        """
        query = gql(
            """
            mutation CommitCreate ($commit: CommitCreateInput!)
                { commitCreate(commit: $commit)}
            """
        )
        params = {
            "commit": {
                "streamId": stream_id,
                "branchName": branch_name,
                "objectId": object_id,
                "message": message,
                "sourceApplication": source_application,
            }
        }
        if parents:
            params["commit"]["parents"] = parents

        return self.make_request(
            query=query, params=params, return_type="commitCreate", parse_response=False
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def update(self, stream_id: str, commit_id: str, message: str) -> bool:
        """
        Update a commit

        Arguments:
            stream_id {str}
                -- the id of the stream that contains the commit you'd like to update
            commit_id {str} -- the id of the commit you'd like to update
            message {str} -- the updated commit message

        Returns:
            bool -- True if the operation succeeded
        """
        query = gql(
            """
            mutation CommitUpdate($commit: CommitUpdateInput!)
                { commitUpdate(commit: $commit)}
            """
        )
        params = {
            "commit": {"streamId": stream_id, "id": commit_id, "message": message}
        }

        return self.make_request(
            query=query, params=params, return_type="commitUpdate", parse_response=False
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def delete(self, stream_id: str, commit_id: str) -> bool:
        """
        Delete a commit

        Arguments:
            stream_id {str}
                -- the id of the stream that contains the commit you'd like to delete
            commit_id {str} -- the id of the commit you'd like to delete

        Returns:
            bool -- True if the operation succeeded
        """
        query = gql(
            """
            mutation CommitDelete($commit: CommitDeleteInput!)
                { commitDelete(commit: $commit)}
            """
        )
        params = {"commit": {"streamId": stream_id, "id": commit_id}}

        return self.make_request(
            query=query, params=params, return_type="commitDelete", parse_response=False
        )

    @deprecated(reason=FE1_DEPRECATION_REASON, version=FE1_DEPRECATION_VERSION)
    def received(
        self,
        stream_id: str,
        commit_id: str,
        source_application: str = "python",
        message: Optional[str] = None,
    ) -> bool:
        """
        Mark a commit object a received by the source application.
        """
        query = gql(
            """
            mutation CommitReceive($receivedInput:CommitReceivedInput!){
                commitReceive(input:$receivedInput)
            }
            """
        )
        params = {
            "receivedInput": {
                "sourceApplication": source_application,
                "streamId": stream_id,
                "commitId": commit_id,
                "message": "message",
            }
        }

        try:
            return self.make_request(
                query=query,
                params=params,
                return_type="commitReceive",
                parse_response=False,
            )
        except Exception as ex:
            print(ex.with_traceback)
            return False



from typing import Optional

from specklepy.objects.base import Base


class CRS(Base, speckle_type="Objects.GIS.CRS"):
    """A Coordinate Reference System stored in wkt format"""

    name: Optional[str] = None
    authority_id: Optional[str] = None
    wkt: Optional[str] = None
    units_native: Optional[str] = None
    offset_x: Optional[float] = None
    offset_y: Optional[float] = None
    rotation: Optional[float] = None



from typing import Optional

from specklepy.core.api.inputs.model_inputs import ModelVersionsFilter
from specklepy.core.api.inputs.version_inputs import (
    CreateVersionInput,
    DeleteVersionsInput,
    MarkReceivedVersionInput,
    MoveVersionsInput,
    UpdateVersionInput,
)
from specklepy.core.api.models import ResourceCollection, Version
from specklepy.core.api.resources import VersionResource as CoreResource
from specklepy.logging import metrics


class VersionResource(CoreResource):
    """API Access class for model versions"""

    def __init__(self, account, basepath, client, server_version) -> None:
        super().__init__(
            account=account,
            basepath=basepath,
            client=client,
            server_version=server_version,
        )

    def get(self, version_id: str, project_id: str) -> Version:
        metrics.track(metrics.SDK, self.account, {"name": "Version Get"})
        return super().get(version_id, project_id)

    def get_versions(
        self,
        model_id: str,
        project_id: str,
        *,
        limit: int = 25,
        cursor: Optional[str] = None,
        filter: Optional[ModelVersionsFilter] = None,
    ) -> ResourceCollection[Version]:
        metrics.track(metrics.SDK, self.account, {"name": "Version Get Versions"})
        return super().get_versions(
            model_id, project_id, limit=limit, cursor=cursor, filter=filter
        )

    def create(self, input: CreateVersionInput) -> str:
        metrics.track(metrics.SDK, self.account, {"name": "Version Create"})
        return super().create(input)

    def update(self, input: UpdateVersionInput) -> Version:
        metrics.track(metrics.SDK, self.account, {"name": "Version Update"})
        return super().update(input)

    def move_to_model(self, input: MoveVersionsInput) -> str:
        metrics.track(metrics.SDK, self.account, {"name": "Version Move To Model"})
        return super().move_to_model(input)

    def delete(self, input: DeleteVersionsInput) -> bool:
        metrics.track(metrics.SDK, self.account, {"name": "Version Delete"})
        return super().delete(input)

    def received(self, input: MarkReceivedVersionInput) -> bool:
        metrics.track(metrics.SDK, self.account, {"name": "Version Received"})
        return super().received(input)



from typing import Optional, Sequence

from pydantic import BaseModel


class UpdateVersionInput(BaseModel):
    versionId: str
    projectId: str
    message: Optional[str]


class MoveVersionsInput(BaseModel):
    targetModelName: str
    versionIds: Sequence[str]
    projectId: str


class DeleteVersionsInput(BaseModel):
    versionIds: Sequence[str]
    projectId: str


class CreateVersionInput(BaseModel):
    objectId: str
    modelId: str
    projectId: str
    message: Optional[str] = None
    sourceApplication: Optional[str] = "py"
    totalChildrenCount: Optional[int] = None
    parents: Optional[Sequence[str]] = None


class MarkReceivedVersionInput(BaseModel):
    versionId: str
    projectId: str
    sourceApplication: str
    message: Optional[str] = None



import gzip
import json
import logging
import queue
import threading

import requests

from specklepy.logging.exceptions import SpeckleException

LOG = logging.getLogger(__name__)


class BatchSender(object):
    def __init__(
        self,
        server_url,
        stream_id,
        token,
        max_batch_size_mb=1,
        max_batch_length=20000,
        batch_buffer_length=10,
        thread_count=4,
    ):
        self.server_url = server_url
        self.stream_id = stream_id
        self._token = token

        self.max_size = int(max_batch_size_mb * 1000 * 1000)
        self.max_batch_length = int(max_batch_length)
        self._batches = queue.Queue(batch_buffer_length)
        self._crt_batch = []
        self._crt_batch_size = 0

        self.thread_count = thread_count
        self._send_threads = []
        self._exception = None

    def send_object(self, id: str, obj: str):
        if not self._send_threads:
            self._create_threads()

        crt_obj_size = len(obj)
        crt_batch_length = len(self._crt_batch)
        if not self._crt_batch or (
            self._crt_batch_size + crt_obj_size < self.max_size
            and crt_batch_length < self.max_batch_length
        ):
            self._crt_batch.append((id, obj))
            self._crt_batch_size += crt_obj_size
            return

        self._batches.put(self._crt_batch)
        self._crt_batch = [(id, obj)]
        self._crt_batch_size = crt_obj_size

    def flush(self):
        # Add current non-complete batch
        if self._crt_batch:
            self._batches.put(self._crt_batch)
            self._crt_batch = []
            self._crt_batch_size = 0
        # Wait for queued batches to be sent
        self._batches.join()
        # End the sending threads
        self._delete_threads()
        # If there was any error, throw the first exception that occurred during upload
        if self._exception is not None:
            ex = self._exception
            self._exception = None
            raise ex

    def _sending_thread_main(self):
        try:
            session = requests.Session()
            session.headers.update(
                {"Authorization": f"Bearer {self._token}", "Accept": "text/plain"}
            )

            while True:
                batch = self._batches.get()

                # None is a sentinel value, meaning the thread should exit gracefully
                if batch is None:
                    self._batches.task_done()
                    break

                try:
                    self._bg_send_batch(session, batch)
                except Exception as ex:
                    self._exception = self._exception or ex
                    LOG.error("Error sending batch of objects to server: " + str(ex))

                self._batches.task_done()
        except Exception as ex:
            self._exception = self._exception or ex
            LOG.error("ServerTransport sending thread error: " + str(ex))

    def _bg_send_batch(self, session: requests.Session, batch):
        object_ids = [obj[0] for obj in batch]
        response = session.post(
            url=f"{self.server_url}/api/diff/{self.stream_id}",
            data={"objects": json.dumps(object_ids)},
        )
        if response.status_code == 403:
            raise SpeckleException(
                f"Invalid credentials - cannot send objects to server {self.server_url}"
            )
        response.raise_for_status()
        server_has_object = response.json()

        new_object_ids = [x for x in object_ids if not server_has_object[x]]
        new_object_ids = set(new_object_ids)
        new_objects = [obj[1] for obj in batch if obj[0] in new_object_ids]

        if not new_objects:
            LOG.info(
                f"Uploading batch of {len(batch)} objects: all objects are already in"
                " the server"
            )
            return

        upload_data = "[" + ",".join(new_objects) + "]"
        upload_data_gzip = gzip.compress(upload_data.encode())
        LOG.info(
            "Uploading batch of %s objects (%s new): (size: %s, compressed size: %s)"
            % (len(batch), len(new_objects), len(upload_data), len(upload_data_gzip))
        )

        try:
            r = session.post(
                url=f"{self.server_url}/objects/{self.stream_id}",
                files={"batch-1": ("batch-1", upload_data_gzip, "application/gzip")},
            )
            if r.status_code != 201:
                LOG.warning("Upload server response: %s", r.text)
                raise SpeckleException(
                    message=(
                        "Could not save the object to the server - status code"
                        f" {r.status_code} ({r.text[:1000]})"
                    )
                )
        except json.JSONDecodeError as error:
            return SpeckleException(
                f"Failed to send objects to {self.server_url}. Please ensure this"
                f" stream ({self.stream_id}) exists on this server and that you have"
                " permission to send to it.",
                error,
            )

    def _create_threads(self):
        for _ in range(self.thread_count):
            t = threading.Thread(target=self._sending_thread_main, daemon=True)
            t.start()
            self._send_threads.append(t)

    def _delete_threads(self):
        for _ in range(len(self._send_threads)):
            self._batches.put(None)

        for thread in self._send_threads:
            thread.join()

        self._send_threads = []

    def __del__(self):
        self._delete_threads()
